import numpy as np
import pandas as pd
import warnings

from scipy.spatial import cKDTree
from scipy.stats import moment
from sklearn.preprocessing import MinMaxScaler

# detect if GPU is available to calculate density
try:
    import torch
    G_RAP_GPU = torch.cuda.is_available()
    if G_RAP_GPU:
        import cuml
except ModuleNotFoundError:
    G_RAP_GPU = False

pd.set_option('mode.chained_assignment', None)

def scale_values(data: pd.DataFrame):# -> pd.DataFrame:
    """Scales each column of the given `pandas.DataFrame` to range [0, 1).

    Args:
        data (pandas.DataFrame): A numeric (int/float) matrix.

    Returns:
        pandas.DataFrame: The scaled version of the given matrix.
    """
    ds = data.apply(lambda x: np.squeeze(np.abs(MinMaxScaler().fit_transform(
        np.array(x - 1).reshape((-1, 1)) ))), axis=0)
    return ds

## feature functions ####
def get_bin_density(
    data: np.ndarray | pd.DataFrame,
    bins: list | np.ndarray | pd.Series,
    dens_agg_type: str = 'max',
    n_cores: int = -1,
    p: int = 2, # see kdtree.query
    eps: float = 0.1, # see kdtree.query
    dens_k_dtm: int = 15
):# -> np.ndarray:
    """Calculates a density value for each bin.

    Args:
        data (numpy.ndarray | pandas.DataFrame): Numeric matrix (2D array).
        bins (list | numpy.ndarray | pandas.Series):
            list containing bin value for each `data` row.
        dens_agg_type (str, optional): Aggregation method per bin.
            See pandas.DataFrame.agg(). Defaults to 'max'.
        n_cores (int, optional): The number of cores to use while
            calculating the density feature. Set to -1 to use all cores.
            Defaults to -1.
        p (int, optional): See scipy.spatial.cKDTree.query(). Defaults to 2.
        eps (float, optional): See scipy.spatial.cKDTree.query().
            Defaults to 0.1.
        dens_k_dtm (int, optional): See scipy.spatial.cKDTree.query().
            Defaults to 15.

    Returns:
        numpy.ndarray: A 1D array containing a density value for each bin.
    """
    if len(bins) != len(data):
        raise ValueError('bins and data do not have the same length.')

    # estimate density for each bin
    if G_RAP_GPU:
        model = cuml.neighbors.NearestNeighbors(n_neighbors=dens_k_dtm, p=2)
        model = model.fit(data)
        dd, _ = model.kneighbors(data)
    else:
        kdtree = cKDTree(data)
        dd, _ = kdtree.query(data, k=dens_k_dtm, p=p, workers=n_cores, eps=eps)

    de = (dd ** data.shape[1]).sum(-1) # TODO: comment
    with warnings.catch_warnings():
        # known divide by zero warning, rare
        warnings.simplefilter("ignore")
        d = -np.log(de)
    d[(de == 0) | (de < 0)] = 0
    d[np.isinf(de)] = np.max(d[~np.isinf(d)])
    d = np.array(pd.DataFrame({'d': d, 'bin': bins}).groupby(
        'bin', as_index=False)['d'].agg(dens_agg_type)['d'])

    return d

def get_bin_moments(
    data: np.ndarray | pd.DataFrame,
    bins: list | np.ndarray | pd.Series,
    mmt: int = 2
):# -> np.ndarray:
    """Calculates the `mmt`th moment of each bin.

    Args:
        data (numpy.ndarray | pandas.DataFrame): Numeric matrix (2D array).
            If data has more than one column, we take the sum of each row.
        bins (list | np.ndarray | pd.Series):
            list containing bin value for each `data` row.
        mmt (int, optional): Moment. Defaults to 2.

    Returns:
        numpy.ndarray: A 1D array containing the `mmt`th moment for each bin.
    """
    # estimate variance/skew for each bin
    data = data if data.shape[1] == 1 else np.sum(data, axis=1)
    _, inds = np.unique(bins, return_index=True)
    dv = np.squeeze(np.array(data))
    bsp = np.array(np.split(dv, inds)[1:], dtype=object)
    if len(bsp.shape) == 2: # i.e. all bins same size; work around
        dv = np.concatenate((dv, np.array([0.0], dtype=np.dtype(dv[0]))))
        bsp = np.array(np.split(dv, inds)[1:], dtype=object)
        bsp[-1] = bsp[-1][:-1]
    with warnings.catch_warnings():
        # known catastrophic cancellation warning; values replaced
        warnings.simplefilter("ignore")
        v = np.vectorize(lambda x: np.sum(moment(x, moment=mmt, axis=0)))(bsp)

    # sum and remove infinite values
    not_inf = ~np.isinf(v)
    if np.sum(not_inf) < len(v):
        v[v == np.Infinity] = np.max(v[not_inf])
        v[v == -np.Infinity] = np.min(v[not_inf])

    return v





import numpy as np
import pandas as pd
import warnings

from difflib import get_close_matches
from .utils import is_monotonic, seq_array

## channel functions ####
def get_clean_time_chan(
    data: pd.DataFrame,
    time_chan: str = 'time',
    min_bin_size: int = 2000
):# -> tuple:
    """Finds the time channel in data; if there is no time channel, creates one.

    Args:
        data (pandas.DataFrame): _description_
        time_chan (str, optional): _description_. Defaults to 'time'.
        min_bin_size (int, optional): _description_. Defaults to 2000.

    Returns:
        tuple:
            str: Time channel column name in data.
            pandas.DataFrame: The given data sorted by the time channel.
    """
    if time_chan in data.columns:
        tc = time_chan
    else:
        tc = [tci for tci in data.columns if tci.lower() == time_chan.lower()]
        tc_none = len(tc) == 0
        if len(tc) == 1:
            tc = tc[0]
        elif len(tc) > 1:
            tc = get_close_matches(time_chan, tc)[0]
        elif not isinstance(tc, str):
            tc_none = True

        # if time channel doesn't exist, create one
        if tc_none:
            wm = 'No valid time channel in data; proceeding to create one.'
            warnings.warn(wm)
            data[time_chan] = seq_array(size=len(data), max_val=min_bin_size)
            tc = time_chan

    if not is_monotonic(data[tc], strict=False):
        data.sort_values(by=tc, inplace=True)

    # i.e. data.reset_index(drop=False, inplace=True)
    data.index = np.array(range(len(data)))
    return tc, data

def verify_channels(given_chans: list = [], all_chans: list = []):# -> np.ndarray:
    """Verifies whether elements of `given_chans` are in `all_chans`.

    Args:
        given_chans (list): User given strings.
        all_chans (list): The full string list.

    Returns:
        numpy.ndarray: User given strings that are in the full string list.
    """
    if len(given_chans) == 0 or len(all_chans) == 0:
        return []

    gcs = [gc.lower() for gc in given_chans]
    acs = [ac.lower() for ac in all_chans]
    ch_in_data = np.isin(gcs, acs)
    if any(ch_in_data):
        if not all(ch_in_data):
            wm = 'Dropping channels not found in usable data: {}'.format(
                np.array(given_chans)[~ch_in_data])
            warnings.warn(wm)
        gcs = np.array(given_chans)[ch_in_data]
    else:
        warnings.warn('Channels not found in usable data, to find manually.')
        gcs = []

    return np.array(gcs)

def get_clean_fp_channels(
    data: pd.DataFrame,
    fluo_chans: list = [],
    phys_chans: list = [],
    channel_unique_no: int = 25, # omip 018 has very little unique values
    phys_channel_suffix: list = [
        'fs', 'ss', 'area', 'eccentricity', 'forward', 'side'],
    bad_suffix: list = ['bead', 'event', 'label', 'is_gate',
                        'index', 'index_original', 'bin', 'time']
):# -> tuple:
    """Identify fluorescent channels.

    Args:
        data (pandas.DataFrame): FCS pandas.DataFrame.
        fluo_chans (list): String vector containing fluorescent channel names.
            Defaults to None.
        phys_chans (list): String vector containing physical morphology
            channel names. Defaults to None.
        channel_unique_no (int): Minimum number of rows in each column that
            can be non-unique. Defaults to 25.
        phys_channel_suffix (list): Standard suffixes for phys_chans.
            Defaults to ['fs', 'ss', 'area', 'eccentricity', 'forward', 'side'].
        bad_suffix (list): Suffixes to avoid. Defaults to ['bead', 'event',
            'label', 'is_gate', 'index', 'index_original', 'bin', 'time'].

    Returns:
        tuple:
            numpy.array: fluorescent channels that can be used to clean FCS data.
            numpy.array: physical morphology channels.
    """
    # validate all columns
    cc = data.columns
    cc = [f for f in cc if not np.any([
        f.lower().startswith(bs) for bs in bad_suffix])]
    if len(cc) == 0:
        raise ValueError('No valid channels found in data.')
    cc = [c for c in cc if not (is_monotonic(np.array(data[c]), strict=False))]
    if len(cc) == 0:
        raise ValueError('All channels are monotonically in/decreasing.')
    channel_unique_no = min(channel_unique_no, int(len(data) / 2))
    cc = [c for c in cc if len(np.unique(data[c])) > channel_unique_no]
    if len(cc) == 0:
        em = 'All channels have less than {} unique values.'.format(
            channel_unique_no)
        raise ValueError(em)

    # verify user given physical channels
    pc = verify_channels(given_chans=phys_chans, all_chans=cc)
    if len(pc) == 0:
        pc = [f for f in cc if np.any(
            [f.lower().startswith(ps) for ps in phys_channel_suffix])]

    # verify user given fluorescent channels
    fc = verify_channels(given_chans=fluo_chans, all_chans=cc)
    if len(fc) == 0:
        fc1 = cc if len(pc) == 0 else [f for f in cc if not (f in pc)]
        if len(fc1) > 0:
            fc2 = [f for f in fc1 if f.lower().endswith('-a')]
            if len(fc2) == 0:
                fc2 = fc1
            fc = fc2 if len(fc2) >= len(fc) else fc

    # verify user given physical channels part 2
    if len(pc) > 0:
        pc2 = [p for p in pc if p.lower().endswith('-a')]
        if len(pc2) == 0:
            pc2 = pc
        pc = pc2 if len(pc2) <= len(pc) and len(pc2) > 0 else pc

    return np.array(fc), np.array(pc)

def most_corr_channels(
    data: pd.DataFrame,
    chosen_chans: np.ndarray | pd.Series | None = None,
    bins: np.ndarray | pd.Series | None = None,
    candidate_no: int = 4,
    min_nrows: int = 50,
    corr_type: str = 'max'
):# -> np.ndarray:
    """Verifies chosen channels and finds channels most correlated with time.

    Args:
        data (pd.DataFrame): FCS data matrix.
        chosen_chans (np.ndarray | pd.Series | None, optional):
            User chosen channels. Defaults to None.
        bins (np.ndarray | pd.Series | None, optional): Bin labels.
            Defaults to None.
        candidate_no (int, optional): Number of channels to return.
            Defaults to 4.
        min_nrows (int, optional): Minimum number of rows required to calculate
            correlation. Defaults to 50.
        corr_type (str, optional): Type of summarization to use on bins to
            calculate correlation e.g. `min`, `max`, `median`, `mean`.
            See `pandas.DataFrame.agg`. Defaults to 'max'.

    Returns:
        np.ndarray: Vector of candidate channel names.
    """
    # check if user chosen channels are in data
    if not (chosen_chans is None):
        corr_ch_in_data = np.isin(chosen_chans, data.columns)
        if any(corr_ch_in_data):
            if not all(corr_ch_in_data):
                wm = 'Dropping channels not found in data: {}'.format(
                    chosen_chans[~corr_ch_in_data])
                warnings.warn(wm)
            return chosen_chans[corr_ch_in_data]
        warnings.warn('Chosen channels not found in data, will find manually.')

    # check edge cases
    candidate_chans = np.array(data.columns)
    if len(data) <= min_nrows:
        warnings.warn('Data has too little rows, returning random channels.')
    if len(candidate_chans) <= candidate_no:
        return candidate_chans
    if len(data) <= min_nrows:
        return candidate_chans[:candidate_no]

    # calculate correlation
    bins = np.array(range(len(data))) if bins is None else bins
    corr_bin = data.groupby(bins).agg(corr_type)
    _, time_binned_nb = np.unique(bins, return_index=True)
    # correlation with bin rate
    # time_binned_nb = np.diff(np.concatenate((inds, [len(data)])))
    corrs = np.zeros((len(candidate_chans)))
    for ci in range(len(candidate_chans)):
        chan = candidate_chans[ci]
        if len(np.unique(corr_bin[chan])) == 1:
            corrs[ci] = 0.0
        else:
            corrs[ci] = np.corrcoef(time_binned_nb, corr_bin[chan])[0, 1]

    return np.array(candidate_chans)[np.argsort(0-np.abs(corrs))[:candidate_no]]




import numpy as np
import pandas as pd
import datetime
import warnings

from .utils import str_to_time, round_to_1

## binning functions ####
def get_time_bin_unit(
    time_values: list | np.ndarray | pd.Series,
    time_bin: str | int | float | None = None,
    bin_size_temp: int = 1000000,
    bin_no_max_temp: int = 10
):# -> str:
    """Determine a relative time bin unit.

    Args:
        time_values (list | numpy.ndarray | pandas.Series): Time channel values.
        time_bin (str | int | float | None, optional): The desired time bin
            size, in seconds or milliseconds. Defaults to None.
        bin_size_temp (int, optional): Default bin frequency.
            Defaults to 1000000.
        bin_no_max_temp (int, optional): Maximum bin size. Defaults to 10.

    Returns:
        str: Time bin unit.
    """
    if isinstance(time_bin, str):
        return time_bin

    c = round_to_1(min(int(np.ceil(
        max(time_values) / bin_size_temp)), bin_no_max_temp))
    if c == 1:
        tb = 'S'
    elif c < 0.001:
        tb = str(c * 1000) + 'ms'
    else:
        tb = str(c) + 'S'
    return tb

def adjust_time_bin(
    time_bin: str,
    frac: float = 0.5,
    reduce: bool = True
):# -> tuple:
    """Reduce a time bin size by a fraction and return the new time bin size and scale.

    Args:
        time_bin (str): The initial time bin size.
        frac (int, optional): The fraction to reduce the time bin size by.
            Defaults to 2.
        Reduce (bool, optional): whether to reduce or increase time bin.
            Defaults to True.

    Returns:
        tuple:
            str: time_bin.
            str: time_unit.
    """
    tb, time_unit = str_to_time(time_bin)
    new_tb = tb * (frac if reduce else (1/frac))
    if time_unit == 'S' and new_tb < 0.001 and reduce:
            time_unit = 'ms'
            new_tb = new_tb * 1000
    elif time_unit == 'ms':
        if reduce and new_tb < 0.001:
            time_unit = 'us'
            new_tb = new_tb * 1000
        elif not reduce and new_tb >= 1000:
            time_unit = 'S'
            new_tb = new_tb / 1000
    elif time_unit == 'us' and new_tb >= 1000 and not reduce:
        time_unit = 'ms'
        new_tb = new_tb / 1000

    time_bin = str(round_to_1(new_tb)) + time_unit
    return time_bin, time_unit

def data_to_flow(
    time_values: np.ndarray | pd.Series,
    min_bin: int = 2000,
    max_time: int = int(1e9), # 1 second
    min_time: int = 1000  # TODO: ???
):# -> pd.DataFrame:
    """Convert a dataset with timestamps into a flow dataframe.

    Args:
        time_values (numpy.ndarray | pandas.Series): Time channel values.
        min_bin (int, optional): Minimum number of bins. Defaults to 2000.
        max_time (int, optional): Maximum time scale to use.
            Defaults to 1e9 (1 second).
        min_time (int, optional): Maximum time scale to use. Defaults to 1000.

    Returns:
        pandas.DataFrame: A flow dataframe with timestamped rows
            and an 'index' column.
    """
    max_t = max(time_values)
    if max_t > max_time:
        type_dt = 'timedelta64[us]'
    elif max_t < min_time:
        type_dt = 'timedelta64[s]'
    else:
        type_dt = 'timedelta64[ms]'

    now = np.datetime64(datetime.datetime(1970, 1, 1)) + np.array(
        time_values, dtype=type_dt)
    flow = pd.DataFrame({'index': np.array(range(len(time_values)))})
    flow.index = pd.DatetimeIndex(now)

    # make time points unique
    time_point, inds, n = np.unique(
        flow.index, return_index=True, return_counts=True)
    if len(time_point) < min_bin:
        if len(time_point) == 1:
            tp = np.array(flow.index) + (max_time * np.array(range(len(flow))))
        else:
            time_interval = time_point[1] - time_point[0]
            time_eps = [time_interval / n[i] for i in range(len(time_point)-1)]
            time_eps.append(np.min(time_eps))
            tpl = [np.array(flow.index[inds[i]:(inds[i] + n[i])]) + np.cumsum(
                np.full((n[i]), time_eps[i])) for i in range(len(time_point))]
            tp = np.concatenate(tpl)

        flow.index = tp

    return flow

def _get_time_bin_while(
    flow_df: pd.DataFrame,
    time_bin: str,
    reduce: bool
):# -> tuple:
    nb = np.array(flow_df.resample(time_bin).agg({"index": "size"}))
    mask = nb > 0
    nb = nb[mask]
    last_time = time_bin
    time_bin, time_unit = adjust_time_bin(time_bin, reduce=reduce)
    return last_time, time_bin, time_unit, nb, mask

def get_time_binned(
    time_values: np.ndarray | pd.Series,
    time_bin: str = '1S',
    time_step: float | None = None,
    min_bin_size: int = 2000,
    max_bin_size: int = 10000,
    min_events_per_bin: int = 50
):# -> tuple:
    """Bin the time channel values.

    Args:
        time_values (numpy.ndarray | pandas.Series): Time channel values.
        time_bin (str, optional): Time bin i.e. duration of each bin.
            Defaults to '1S'.
        time_step (float | None, optional): Time step parameter from FCS file
            e.g. `meta.get('$TIMESTEP')`. Defaults to None.
        min_bin_size (int, optional): Minimum number of bins. Defaults to 2000.
        max_bin_size (int, optional): Maximum number of bins. Defaults to 10000.
        min_events_per_bin (int, optional): Minimum number of events per bin.
            Defaults to 50.

    Returns:
        numpy.ndarray: A 1D numpy array with the same length as `time_values`
            containing integer bin labels.
    """
    if not (time_step is None):
        # count number of events grouped by time
        bin_no = int(len(time_values) / min_events_per_bin)
        min_bin_size = min(min_bin_size, bin_no)
        flow_df = data_to_flow(time_values=time_values, min_bin=min_bin_size)

        # iteratively reduce/increase time bin
        nb = np.array(flow_df.resample(time_bin).agg({"index": "size"}))
        mask = nb > 0
        nb = nb[mask]
        last_time = time_bin
        _, time_unit = str_to_time(time_bin)
        if len(nb) < min_bin_size and time_unit != 'us':
            while len(nb) < min_bin_size and time_unit != 'us':
                last_time, time_bin, time_unit, nb, mask = _get_time_bin_while(
                    flow_df, time_bin, reduce=True)
        elif len(nb) > max_bin_size:
            while len(nb) > max_bin_size:
                last_time, time_bin, time_unit, nb, mask = _get_time_bin_while(
                    flow_df, time_bin, reduce=False)

        if last_time is None:
            em = 'data has too few values: ({})'.format(len(time_values))
            raise ValueError(em)

        time_bin = last_time
        index = np.array(flow_df.resample(time_bin).agg({"index": "max"}))[mask]
        bin_r = np.digitize(
            np.array(flow_df["index"]), np.append([0], index), right=True)
        bin_r[0] = 1
    else:
        wm = '3.X fcs required to compute extreme studentized deviate test.'
        warnings.warn(wm)

        # default time_step = 1.0
        nbin = min(min_bin_size, len(time_values) / min_events_per_bin)
        bins = np.linspace(0, len(time_values) - 1, int(nbin))
        bin_r = np.digitize(np.array(range(len(time_values))), bins)

    return bin_r





import numpy as np
import warnings
import pathlib
import matplotlib.pyplot as plt

## plotting functions ###
def plot_scat(
    x, y, L = None,
    plot_title: str = 'Scatterplot',
    fig_size: tuple = (16, 8),
    dot_size: float = 3,
    out_path: str = str(pathlib.Path().resolve()) + '/temp.png'
):# -> None:
    """Plots and saves a scatterplot.

    Args:
        x (list | numpy.array | pandas.Series): x axis values.
        y (list | numpy.array | pandas.Series): y axis values.
        L (list | numpy.array | pandas.Series | None, optional): Category label for
            each value. If set to `None`, there will be only one category.
            Defaults to None.
        plot_title (str, optional): Plot title. Defaults to 'Scatterplot'.
        fig_size (tuple, optional): Plot width and height. Defaults to (16, 8).
        dot_size (float, optional): Scatterplot point/dot size. Defaults to 3.
        out_path (str, optional): Path to save plot.
            Defaults to str(pathlib.Path().resolve())+'/temp.png'.

    Returns:
        None: Function does not return anything, it saves plot to file.
    """
    # prepare arguments
    if len(x) != len(y):
        raise ValueError('x is not the same length as y: ', len(x), ' vs ', len(y))
    if len(x) == 0:
        warnings.warn('x contains no values, ignoring plot.')
        return None
    if L is None or len(x) != len(x):
        warnings.warn('plot colour categories maldefined, using single colour.')
        L = np.zeros((len(x)))

    x = np.array(x)
    y = np.array(y)
    L = np.array(L)

    # plot
    _, ax = plt.subplots(figsize=fig_size)
    Ls = np.unique(L)
    for l in Ls:
        mask = (L == l)
        ax.scatter(x[mask], y[mask], label=l, s=dot_size)
    ax.set_title(plot_title)
    if len(Ls) > 1:
        ax.legend()

    # save plot
    plt.savefig(out_path)
    plt.close()




from functools import partial
from typing import Type
import time
import copy
import warnings
import pandas as pd
import numpy as np
import ruptures as rpt
from sklearn.preprocessing import RobustScaler
from sklearn.metrics import silhouette_score, pairwise_distances

from metaclean3.utils import mode, order_value_bins, randomize_duplicates, arg_lim
from metaclean3.plot import plot_scat
from metaclean3.segments import (
    in_percentiles,
    go_updown,
    get_ref_updown,
    remove_reference_ends,
    merge_segments_test,
    merge_segments_gain,
    merge_segments_quant,
    refine_reference_segment,
    refine_reference_segment_inbetween
)
from metaclean3.changepoint import chpts0M_to_segments, segments_to_chpts0M, list_gains
from metaclean3.outliers import IsoForestDetector, drop_outliers
from metaclean3.features import get_bin_density, get_bin_moments, scale_values
from metaclean3.fcs import FCSfile
from metaclean3.channels import most_corr_channels

class MetaClean:
    """Cleans a given data set.
    """
    def __init__(
        self,
        # for rupture: changepoint
        seg_method: str = 'pelt',
        cost_model_seg: str = 'rbf',
        jump_per_seg: int = 2,
        min_seg_size: int = 2,
        # for binseg
        seg_no: int = 40, # over-estimate segments please > 0
        mean_no_per_seg: int = 50,
        min_no_per_seg_limit: int = 10,
        # for pelt
        pelt_penalty: float = 5.0,
        # for combining segments (significance test) / determining best segment
        merge_method: str = 'sequential',
        merge_signif_test: str = 'wilcox', # wilcoxan, ...
        p_thres: float = 0.05, # > 0, < 1
        signif_strict: bool = True,
        # for combining segments (gain)
        min_ref_percent: float = 0.5,
        cost_model_gain: str = 'rank',
        gain_diff: bool = True, # ratio if false
        # for combining segments (quantile ranges)
        percent_diff: float = 0.05,
        percent_shifts: float = [0.15, 0.2, 0.25, 0.3, 0.35, 0.4],
        small_no_per_seg0: int = 50,
        ignore_no: int = 10,
        ref_percents: list = [1, 0.5],
        ref_percent: float = 0.1,
        exception_no: int = 2,
        keep_skipped: str = 'all',
        # be lenient when ordering segments
        pd_lenient: int = 2,
        # if there are non-reference segments that are large, keep it
        min_ref_percent_to_keep: float = 0.4,
        # output options
        random_seed: int = 623,
        verbose: bool = True,
        png_dir: str = ''
    ):
        """
        Args:
            seg_method (str, optional): Changepoint detection method; currently
                support options 'pelt' and 'binseg'. Defaults to 'pelt'.
            cost_model_seg (str, optional): See cost models in `ruptures.costs`.
                Used for finding changepoints. Defaults to 'rbf'.
            jump_per_seg (int, optional): See `jump` argument in
                `ruptures.Binseg`. Defaults to 2.
            min_seg_size (int, optional): See `min_size` argument in
                `ruptures.Binseg`. Defaults to 2.
            seg_no (int, optional): Number of segments to find; see `n_bkps`
                argument in `ruptures.base` (used only if `seg_method`
                is set to `binseg`). Defaults to 40.
            min_no_per_seg_limit (int, optional): The maximum length of a
                "small" segment; small segments are treated more leniently when
                deciding whether or not it should be merged with adjacent
                segments. Defaults to 10.
            pelt_penalty (float, optional): See `pen` argument in
                `ruptures.KernelCPD`. Defaults to 5.0.
            merge_method (str, optional): 'sequential' or 'sd'. 'sequential'
                merges adjacent segments that are not significantly different.
                'sd' removes changepoints that have a gain within
                3 standard deviations. Defaults to 'sequential'.
            merge_signif_test (str, optional): 'wilcox' or 't'.
                Significance test to use for testing and merging segments.
                Defaults to 'wilcox'.
            min_ref_percent (float, optional): Stopping criteria; indicates the
                minimum length (proportion of total) of any segment.
                Defaults to 0.5.
            cost_model_gain (str, optional): See cost models in `ruptures.costs`.
                Used for calculating changepoint gain. Defaults to 'rank'.
            gain_diff (bool, optional): Wether to use the difference (`True`)
                or the ratio (`False`) to calculate gain change.
                Defaults to True.
            percent_shifts (list, optional): Quantile ranges used to test
                two segments' quantile range.
                Defaults to [0.15, 0.2, 0.25, 0.3, 0.35, 0.4].
            small_no_per_seg0 (int, optional): An upper limit for and used to
                refine `min_no_per_seg_limit`. Defaults to 50.
            ignore_no (int, optional): The maximum length of a segment that has
                negligible length. Defaults to 10.
            ref_percents (list, optional): The size of reference segments to
                test (a proportion of the reference segment length).
                Defaults to [1, 0.5].
            ref_percent (float, optional): The size of reference segments to
                test (a proportion of the whole length). Defaults to 0.1.
            exception_no (int, optional): Users can choose to check test values
                `exception_no` segments away from reference segment
                (i.e. non-adjacent test segments). Defaults to 2.
            keep_skipped (str, optional): Remove 'all', 'some', or 'none'
                skipped segments. Defaults to 'all'.
            min_ref_percent_to_keep (float, optional): If there are
                non-reference segments larger than this percentage, we keep it.
                Defaults to 0.4.
            pd_lenient (int, optional): How many times as lenient should
                the function be when ordering (not merging) segments
                based on their geometric mean and quantiles. Defaults to 2.
            random_seed (int, optional): Random seed. Defaults to 623.
            verbose (bool, optional): Whether to display progress messages.
                Defaults to True.
            png_dir (str, optional): .png path of where users want to save
                plots; no plot will be made if set to ''. Defaults to ''.
        """
        # segmentation arguments
        # for binseg
        self.seg_no = seg_no # over-estimate segments please > 0
        self.mean_no_per_seg = mean_no_per_seg
        if seg_method not in ['pelt', 'binseg']: # Literal['binseg', 'bayes', 'kmeans2']
            raise ValueError('seg_method chosen is not available.')
        self.seg_method = seg_method

        # for binseg
        bms = [
            'rbf', 'normal', 'linear', 'l1', 'l2', 'ar', 'mi', 'rank', 'cosine']
        if cost_model_seg not in bms:
            raise ValueError('cost_model_seg chosen is not available.')
        if cost_model_gain not in bms:
            raise ValueError('cost_model_gain chosen is not available.')
        self.cost_model_seg = cost_model_seg
        self.cost_model_gain = cost_model_gain
        self.jump_per_seg = arg_lim(jump_per_seg, xmin=1)
        min_no_per_seg_limit = min(
            min_no_per_seg_limit, np.round(small_no_per_seg0 / 2))
        # HARDCODE because segment merging WILL error if less than 5
        self.min_no_per_seg_limit = arg_lim(min_no_per_seg_limit, xmin=5)
        self.min_seg_size = arg_lim(min_seg_size, xmin=1)
        self.pelt_penalty = pelt_penalty
        self.gain_diff = gain_diff # ratio if false

        # combine segment / determine best segment

        if merge_method not in ['sequential', 'sequential']: # Literal['sequential', 'sd']
            raise ValueError('{}{}'.format(
                'merge_method chosen is not available, ',
                'please choose from \'sequential\' or \'sequential\''))
        self.merge_method = merge_method
        if merge_signif_test not in ['t', 'cohen', 'wilcox']: # Literal['t', 'cohen']
            raise ValueError('{}{}'.format(
                'merge_signif_test chosen is not available, ',
                'please choose from \'t\', \'cohen\', or \'wilcox\''))
        self.merge_signif_test = merge_signif_test
        self.p_thres = arg_lim(p_thres, xmin=0, xmax=1)
        self.signif_strict = signif_strict

        if keep_skipped not in ['all', 'some', 'none']: # Literal['all', 'some', 'none']
            raise ValueError('{}{}'.format(
                'merge_signif_test chosen is not available, ',
                'please choose from \'all\', \'some\', or \'none\''))
        self.keep_skipped = keep_skipped
        self.percent_diff = arg_lim(percent_diff, xmin=0, xmax=1)
        self.ignore_no = arg_lim(ignore_no, xmin=0)
        self.percent_shifts = [
            arg_lim(p, xmin=0, xmax=1) for p in percent_shifts]
        self.in_percentiles_ind = 0 # which percent_shifts to use
        self.exception_no = arg_lim(exception_no, xmin=0)

        self.min_ref_percent = arg_lim(min_ref_percent, xmin=0, xmax=1)
        self.small_no_per_seg0 = arg_lim(small_no_per_seg0, xmin=0, xmax=1)
        self.small_no_per_seg = arg_lim(small_no_per_seg0, xmin=0, xmax=1)
        self.ref_percents = [arg_lim(r, xmin=0, xmax=1) for r in ref_percents]
        self.ref_percent = arg_lim(ref_percent, xmin=0, xmax=1)

        # functions
        self.in_percentiles = None
        self.go_updown = None
        self.get_ref_fun = None

        # generic arguments
        self.pd_lenient = arg_lim(pd_lenient, xmin=1)
        self.min_ref_percent_to_keep = arg_lim(
            min_ref_percent_to_keep, xmin=0, xmax=1)
        self.random_seed = random_seed
        self.verbose = verbose
        self.png_dir = png_dir

    # TODO: make methods private after testing
    def clean(
        self,
        data: pd.DataFrame,
        val_cols: list | None = None,
        val_col_final: str | None = None,
    ):# -> tuple:
        """Conduct cleaning of `data` (i.e. remove irregular objects/rows).

        Args:
            data (pandas.DataFrame): An object x feature matrix.
            val_cols (list | None): List of feature column names in `data` to be
                used for cleaning. Defaults to None.
            val_col_final (str | None): A summary feature column name in `data`
                where (e.g. sum of `val_cols` values). `val_cols` and
                `val_col_final` can be overlapping. Defaults to None.

        Returns:
            tuple:
                numpy.ndarray: A 1D boolean array with `True`/`False` labels
                    indicating which rows in `data` to keep/remove.
                numpy.ndarray: A 1D integer array containing refined segment
                    labels.
                numpy.ndarray: A 1D integer array containing raw segment labels.
        """
        start0 = time.time()
        # specify columns to use
        if val_cols is None:
            val_cols = data.columns
        if val_col_final is None:
            if len(val_cols) == 1:
                val_col_final = val_cols[0]
            else:
                val_col_final = '_'.join(val_cols)
                data[val_col_final] = data[val_cols].sum(axis=1)
        # process parameters
        self.small_no_per_seg = min(
            self.small_no_per_seg0,
            int(len(data) / self.min_no_per_seg_limit)
        )
        # prepare functions
        self.in_percentiles_multi = [partial(
            in_percentiles,
            ignore_no=self.ignore_no,
            percent_diff=self.percent_diff,
            percent_shift=p,
            small_no_per_seg=self.small_no_per_seg
        ) for p in self.percent_shifts]
        self.go_updown = partial(
            go_updown,
            exception_no=self.exception_no,
            ignore_no=self.ignore_no
        )
        self.get_ref_fun = partial(
            get_ref_updown,
            ref_percents=self.ref_percents,
            ref_percent=self.ref_percent
        )

        # get initial segments
        start = time.time()
        values = np.array(data[val_col_final])
        vinds = np.array(range(len(values)))
        segments0 = self.get_segments(values=values)
        if self.png_dir:
            op = '{}/0.segments.png'.format(self.png_dir)
            plot_scat(
                x=vinds, y=values, L=segments0,
                plot_title='Raw segments', out_path=op
            )
        if self.verbose:
            print(' + Time to identify segments: {}\'s'.format(
                np.round(time.time() - start, 3)))

        # segments_merged, sorted_index = self.merge_segments(
        start = time.time()
        segments1, _ = self.merge_segments(
            values=values, vlist=data[val_cols], segments=segments0)
        # merge skipped segments into reference
        segments2 = self.refine_segments(
            values=values, segments=segments1, ref_segment=mode(segments1))
        if self.png_dir:
            plot_scat(
                x=vinds, y=values, L=segments2,
                plot_title='Refined segments (Done! Keep the largest segment)',
                out_path='{}/3.refined.png'.format(self.png_dir)
            )
        if self.verbose:
            print(' + Time to merge segments: {}\'s'.format(
                np.round(time.time() - start, 3)))
            print('- Total time used: {}\'s'.format(
                np.round(time.time() - start0, 3)))

        return segments2 == mode(segments2), segments2, segments0

    def get_segments(
        self,
        values: list | np.ndarray | pd.Series,
    ):# -> np.ndarray:
        """
        Given an array, break the array into segments (find changepoints) and
        merge those segments based on significance tests to obtain the initial
        "raw" segments.

        Args:
            values (list | numpy.ndarray | pandas.Series): A 1D array of numeric
                values.

        Returns:
            numpy.ndarray: A 1D integer array containing "raw" segment labels.
        """
        # set up parameters (min 2 points per segment)
        self.seg_no = min(
            max(self.seg_no, 1), int(len(values) / self.min_seg_size))
        if self.seg_no < 2:
            raise ValueError('Given data has < 2 data points')

        if self.seg_method == 'binseg':
            rpt_class = rpt.Binseg(
                model=self.cost_model_seg,
                jump=self.jump_per_seg,
                min_size=self.min_seg_size
            )
            chpts = np.array(
                rpt_class.fit_predict(signal=values, n_bkps=self.seg_no))[:-1]
        elif self.seg_method == 'pelt':
            rpt_class = rpt.KernelCPD(
                kernel=self.cost_model_seg, min_size=self.min_seg_size)
            chpts = np.array(rpt_class.fit_predict(
                signal=values, pen=self.pelt_penalty))[:-1]
        chpts0M = np.concatenate(([0], chpts, [len(values)]))
        rpt_class = rpt.Binseg(
            model=self.cost_model_gain,
            jump=self.jump_per_seg,
            min_size=self.min_seg_size
        ).fit(signal=np.array(values))
        gains = list_gains(
            rpt_class=rpt_class, chpts0M=chpts0M, dif=self.gain_diff)[1:-1]

        return self.merge_segments_raw(values=values, chpts=chpts, gains=gains)

    def merge_segments_raw(
        self,
        values: list | np.ndarray | pd.Series,
        chpts: list | np.ndarray | pd.Series,
        gains: list | np.ndarray | pd.Series
    ):# -> np.ndarray:
        """Merge raw segments using statistical significance tests.

        Args:
            values (list | numpy.ndarray | pandas.Series): A 1D array of numeric
                values.
            chpts (list | numpy.ndarray | pandas.Series): A 1D array of
                changepoints not including 0 and length of `values` as the
                first and last elements.
            gains (list | numpy.ndarray | pandas.Series): List of gains
                corresponding to `chpts`.

        Returns:
            numpy.ndarray: A 1D integer array containing "raw" segment labels.
        """
        if len(chpts) == 0 or (self.merge_method == 'sd' and len(chpts) < 2):
            return np.full((len(values)), 0)

        # merge changepoints
        chpts0M = np.concatenate(([0], chpts, [len(values)]))
        if self.merge_method == 'sd':
            chpts = chpts[np.where(gains > np.mean(gains) + 3 * np.std(gains))]
            chpts0M = np.concatenate(([0], chpts, [len(values)]))
            segments = chpts0M_to_segments(chpts0M)
        elif self.merge_method == 'sequential':
            segments = merge_segments_test(
                values=values,
                segments=chpts0M_to_segments(chpts0M),
                merge_signif_test=self.merge_signif_test,
                p_thres=self.p_thres,
                ignore_no=self.ignore_no,
                strict=self.signif_strict
            )
        # segments = chpts0M_to_segments(np.concatenate(([0], chpts0M[1:-1]+1, [len(values)])))
        return segments

    def merge_segments(
        self,
        values: list | np.ndarray | pd.Series,
        vlist: list,
        segments: list | np.ndarray | pd.Series
    ):# -> tuple:
        """Merge segments based on changepoint gain and segment quantile ranges.

        Args:
            values (list | numpy.ndarray | pandas.Series): A 1D array of numeric
                values.
            vlist (list): A list of 1D arrays with the same length as `values`.
            segments (list | numpy.ndarray | pandas.Series): Segment label
                vector.

        Returns:
            tuple:
                numpy.ndarray: sorted 1D integer segment label array.
                numpy.ndarray: sorted indices.

        """
        # sort segments
        sorted_index = order_value_bins(
            vlist=copy.copy(vlist),
            segments=segments,
            percentile_diff=self.percent_diff * self.pd_lenient
        )

        # merge sorted segments based on gain
        values_sorted = np.array(values)[sorted_index]
        segmentss1 = merge_segments_gain(
            values=values_sorted,
            segments=segments[sorted_index],
            cost_model=self.cost_model_gain,
            binseg_jump=self.jump_per_seg,
            binseg_min_size=self.min_seg_size,
            min_ref_percent=self.min_ref_percent
        )
        if self.png_dir:
            op = '{}/1.merge_segments_gain_SORTED.png'.format(self.png_dir)
            plot_scat(
                x=np.array(range(len(values))), y=values_sorted, L=segmentss1,
                plot_title='Combined segments (gain) SORTED', out_path=op
            )

        # merge sorted segents based on quantiles (best parameter = max sil)
        vlist_sorted = [vlist[v][sorted_index] for v in vlist.columns]
        segmentss2s = [merge_segments_quant(
            vlist=vlist_sorted,
            segments=segmentss1,
            ip_func=partial(
                ip, check_var=False, check_mean=False, switch=True),
            get_ref_fun=self.get_ref_fun,
            small_no_per_seg=self.small_no_per_seg,
            keep_skipped=self.keep_skipped
        ) for ip in self.in_percentiles_multi]
        good_ip = np.full((len(segmentss2s)), True)
        if len(segmentss2s) > 1:
            for i in range(1, len(segmentss2s)):
                if np.array_equal(segmentss2s[i-1], segmentss2s[i]):
                    good_ip[i] = False
        if np.sum(good_ip) > 1:
            good_ip_ind = np.squeeze(np.argwhere(good_ip))
            x = np.expand_dims(values_sorted, axis=1)
            d = pairwise_distances(x)
            sil = [silhouette_score(d, segmentss2s[gip]) if
                   len(np.unique(segmentss2s[gip])) > 1 else 0
                   for gip in good_ip_ind]
            self.in_percentiles_ind = good_ip_ind[np.argmax(sil)]
        segmentss2 = segmentss2s[self.in_percentiles_ind]

        # merge skipped sorted segments into reference
        if self.keep_skipped == 'some' and -1 in segmentss2:
            segmentss2 = refine_reference_segment(
            vlist=[values],
                segments=segmentss2,
                rl=mode(segmentss2),
                sorted_index=sorted_index,
                ip_func=partial(
                    self.in_percentiles_multi[self.in_percentiles_ind],
                    check_mean=False, check_var=True, switch=False),
                small_no_per_seg=self.small_no_per_seg
            )
        if self.png_dir:
            op = '{}/2.merge_segments_quantiles_SORTED.png'.format(self.png_dir)
            plot_scat(
                x=np.array(range(len(values))), y=values_sorted, L=segmentss2,
                plot_title='Combined segments (quantiles) SORTED', out_path=op
            )

        return segmentss2[np.argsort(sorted_index)], sorted_index

    def refine_segments(
        self,
        # vlist: list,
        values: list | np.ndarray | pd.Series,
        segments: list | np.ndarray | pd.Series,
        ref_segment: int
        # sorted_index: list | np.ndarray | None = None
    ):# -> np.ndarray:
        """Refine segment to be kept.

        Args:
            values (list | numpy.ndarray | pandas.Series): A 1D array of numeric
                values.
            segments (list | numpy.ndarray | pandas.Series): Segment label
                vector.
            ref_segment (int): Label of the reference segment.

        Returns:
            numpy.ndarray: refined Segment label vector.
        """
        # segments1 = refine_reference_segment(
        #     vlist=vlist,
        #     segments=copy.copy(segments),
        #     rl=ref_segment,
        #     sorted_index=sorted_index,
        #     ip_func=partial(
        #         self.in_percentiles_multi[self.in_percentiles_ind],
        #         check_var=True, switch=False),
        #     small_no_per_seg=self.small_no_per_seg
        # )
        segments1 = refine_reference_segment_inbetween(
            segments=copy.copy(segments),
            rl=ref_segment,
            small_no_per_seg=self.small_no_per_seg
        )

        # identify large segments
        chpts0M = segments_to_chpts0M(segments1)
        segment_len = np.diff(chpts0M) / len(segments1)
        large_seg = segment_len >= self.min_ref_percent_to_keep
        # identify original reference segment
        ref_chpt = np.squeeze(np.argwhere(segments1 == mode(segments1)))[0]
        large_seg[chpts0M[:-1]==ref_chpt] = True
        # combine the above to create ne reference segment
        large_seg = np.squeeze(np.argwhere(large_seg))
        rl = np.max(segments1) + 1
        if len(large_seg.shape) > 0:
            for i in large_seg:
                segments1[chpts0M[i]:chpts0M[i+1]] = rl

        # remove beginning and ends if applicable
        segments2 = remove_reference_ends(
            values=values,
            segments=copy.copy(segments1),
            rl=mode(segments1),
            ip_func=partial(
                self.in_percentiles_multi[self.in_percentiles_ind],
                check_mean=True, check_var=False, switch=False),
            small_no_per_seg=self.small_no_per_seg * 2
        )

        return segments2

class MetaCleanFCS():
    """
    Given a compensated (flow cytometry) / unmixed (spectral cytometry) and
    transformed cytometry data, prepares the data for cleaning via 0.
    """
    def __init__(
        self,
        # identify best channels for cleaning
        clean_chans: list | np.ndarray | pd.Series | None = None,
        clean_chans_no: int = 4,
        candidate_chans_type: str = 'fluo',
        # corr_method: str = 'spearman',
        corr_type: str = 'max',
        # remove duplicates by row (slow) or by row sums (fast)
        strict_remove_duplicates: bool = False,
        # calculate density
        n_cores: int = -1,
        dens_agg_type: str = 'max', # 'max', 'median'
        dens_k_dtm: int = 15,
        p: int = 2,
        eps: float = 0.1,
        # identify outliers
        outlier_thresh: float = 0.01,
        outlier_trees: int = 500,
        outlier_drop_and: bool = True, # else OR
        rm_outliers: str = 'all',
        # others
        random_seed: int = 623,
        verbose: bool = True,
        png_dir: str = '',
        # metaclean
        metaclean: Type[MetaClean] | None = None,
        **kwargs
    ):
        """
        Args:
            clean_chans (list | numpy.ndarray | pandas.Series | None,
                optional): Users can choose to provide the fluorescent channels
                they want to use. If set to `None`, 0 will select
                channels automatically. Defaults to None.
            clean_chans_no (int, optional): The maximum number of channels
                to use for cleaning. Less channels make feature calculation
                faster and reduces potential noise. Defaults to 4.
            candidate_chans_type (str, optional): The type of channels to use
                for cleaning i.e. 'fluo' (fluorescent),
                'phys' (physical morphology), or 'all'. Defaults to 'fluo'.
            corr_type (str, optional): Type of summarization to use when
                calculating correlation for finding best channels to use for
                cleaning (i.e. 'max', 'mean', or 'min'). Defaults to 'max'.
            strict_remove_duplicates (bool, optional): Whether to remove
                duplicates strictly (by row) (slow) or not (by row sum) (fast).
                Defaults to False.
            n_cores (int, optional): The number of cores to use while
                calculating the density feature. Set to -1 to use all cores.
                Defaults to -1.
            dens_agg_type (str, optional): Type of summarization to use when
                calculating density features (i.e. 'max', 'mean', or 'min').
                Defaults to 'max'.
            p (int, optional): For calculating density features.
                See `kdtree.query`. Defaults to 2.
            eps (float, optional): For calculating density features.
                See `kdtree.query`. Defaults to 0.1.
            outlier_thresh (float, optional): For calculating density features.
                See `contamination` argument in
                `sklearn.ensemble.IsolationForest`. Defaults to 0.01.
            rm_outliers (str, optional): Wether to remove all, some, or none of
                the outliers (i.e. 'all', 'some', or 'none'). Defaults to 'all'.
            random_seed (int, optional): Random seed. Defaults to 623.
            verbose (bool, optional): Whether to display progress messages.
                Defaults to True.
            png_dir (str, optional): .png path of where users want to save
                plots; no plot will be made if set to ''. Defaults to ''.
            metaclean (MetaClean | None, optional): Initialized `MetaClean`
                class. Defaults to None.
            \**kwargs param: See `MetaClean` class attributes.
        """
        self.clean_chans = clean_chans
        self.clean_chans_no = arg_lim(clean_chans_no, xmin=1)
        if corr_type not in ['max', 'mean', 'min']:
            raise ValueError('{}{}'.format(
                'corr_type chosen is not available, ',
                'please choose from \'max\', \'mean\', or \'min\''))
        self.corr_type = corr_type
        if candidate_chans_type not in ['fluo', 'phys', 'all']:
            raise ValueError('{}{}'.format(
                'candidate_chans_type chosen is not available, ',
                'please choose from \'fluo\', \'phys\', or \'all\''))
        self.candidate_chans_type = candidate_chans_type
        self.strict_remove_duplicates = strict_remove_duplicates

        # calculate density
        self.n_cores = 1 if n_cores == 0 else arg_lim(n_cores, xmin=-1)
        if dens_agg_type not in ['max', 'mean', 'min']:
            raise ValueError('{}{}'.format(
                'dens_agg_type chosen is not available, ',
                'please choose from \'max\', \'mean\', or \'min\''))
        self.dens_agg_type = dens_agg_type
        self.p = p # see kdtree.query
        self.eps = eps # see kdtree.query
        self.dens_k_dtm = dens_k_dtm
        self.outlier_func = IsoForestDetector(
            contamination=outlier_thresh,
            n_estimators=outlier_trees,
            random_state=random_seed,
        )
        self.outlier_drop_and = outlier_drop_and
        if rm_outliers not in ['none', 'some', 'all']:
            raise ValueError('{}{}'.format(
                'rm_outliers chosen is not available, ',
                'please choose from \'none\', \'some\', or \'all\''))
        self.rm_outliers = rm_outliers

        self.random_seed = random_seed
        self.verbose = verbose
        self.png_dir = png_dir
        self.fcs = None
        self.metaclean = None

        # fit: metaclean
        if not (metaclean is None):
            self.metaclean = metaclean
        else:
            if 'verbose' not in kwargs:
                kwargs['verbose'] = self.verbose
            if 'png_dir' not in kwargs:
                kwargs['png_dir'] = self.png_dir
            self.metaclean = MetaClean(random_seed=random_seed, **kwargs)

    # apply split in two for testing
    def apply(
        self,
        fcs: Type[FCSfile] | None = None,
        return_binned_data: bool = False,
        **kwargs
    ):# -> pd.DataFrame:
        """Prepare given `fcs` and apply MetaClean.

        Args:
            fcs (FCSfile | None, optional): Initialized `FCSfile` data class.
                Defaults to None.
            return_binned_data (bool, optional): Whether or not to return binned
                data. Defaults to False.
            \**kwargs param: See `FCSfile` class attributes, namely `data`
                (event x featire pandas.DataFrame).

        Returns:
            pandas.DataFrame: A data frame with the same number of rows as
                `fcs.data`. Columns with prefix `val_` contains
                feature values, `outlier_keep` contain boolean values for bins
                considered (`True`) as outliers, `bin` contain bin labels,
                `segments_raw` contain the raw segment labels, `segments`
                contain the merged and refined segment labels, and `clean_keep`
                contain boolean values for bins to keep (`True`) ---
                `clean_keep` is the final recommentation given by 0.
        """
        start0 = time.time()
        data0 = self.apply_features(
            fcs=fcs, calculate_outlier2=self.rm_outliers == 'some', **kwargs)
        data1 = self.apply_clean(data=data0)
        if self.verbose:
            print(' + Total time for MetaCleanFCS: {}\'s'.format(
                np.round(time.time() - start0, 3)))

        # keep outliers if specified
        if self.rm_outliers != 'none':
            dt = self.remove_outliers(copy.copy(data1))
            if not (dt is None):
                data1 = dt

        data_final = pd.merge(self.fcs.data, data1, on=self.fcs.bin_chan)
        data_final.sort_values(by='index_original')

        if return_binned_data:
            return data_final, data1
        return data_final

    def apply_features(
        self,
        fcs: FCSfile | None = None,
        calculate_outlier2: bool = False,
        **kwargs
    ):# -> pd.DataFrame:
        """Calcuate bins and features using raw values.

        Args:
            fcs (FCSfile | None, optional): _description_. Defaults to None.
            calculate_outlier2 (bool, optional): Whether or not to identify
                outliers again more leniently if user wishes to remove only
                some outliers. Defaults to False.

        Returns:
            pandas.DataFrame: A data matrix containing bins (`bin` column),
                feature values (prefix `val_` column(s)), and
                outliers (`outlier_keep` column).
        """
        self.fcs = FCSfile(**kwargs) if fcs is None else fcs
        # find best channels for cleaning
        start = time.time()
        candidate_chans = self.clean_chans
        if candidate_chans is None:
            candidate_chans = np.concatenate(
                (self.fcs.phys_chans, self.fcs.fluo_chans))
            if self.candidate_chans_type == 'fluo':
                if len(self.fcs.fluo_chans) > 0:
                    candidate_chans = self.fcs.fluo_chans
                else:
                    warnings.warn('{}{}'.format(
                        'No fluorescent channels found, ',
                        'defaulting to all channels.'))
            elif self.candidate_chans_type == 'phys':
                if len(self.fcs.phys_chans) > 0:
                    candidate_chans = self.fcs.phys_chans
                else:
                    warnings.warn('{}{}'.format(
                        'No physical morphology channels found, ',
                        'defaulting to all channels.'))
        self.clean_chans = most_corr_channels(
            data=self.fcs.data[candidate_chans],
            bins=self.fcs.data[self.fcs.bin_chan],
            chosen_chans=self.clean_chans,
            candidate_no=self.clean_chans_no,
            corr_type=self.corr_type
        )
        # scale data
        # 210 seconds
        ds_input = randomize_duplicates(
            self.fcs.data[self.clean_chans],
            strict=self.strict_remove_duplicates,
            seed=self.random_seed
        )
        if self.verbose:
            print(' + Time to choose channels and check duplicates: {}\'s ({})'.format(
                np.round(time.time() - start, 3), self.clean_chans))
        start = time.time()
        ds = RobustScaler().fit_transform(ds_input)
        bin_values = self.fcs.data[self.fcs.bin_chan]
        d = get_bin_density( # 29 vs 180 with dups
            data=ds,
            bins=bin_values,
            dens_agg_type=self.dens_agg_type,
            n_cores=self.n_cores,
            p=self.p, # see kdtree.query
            eps=self.eps, # see kdtree.query
            dens_k_dtm=self.dens_k_dtm)
        v = get_bin_moments(data=ds, bins=bin_values, mmt=2)
        s = get_bin_moments(data=ds, bins=bin_values, mmt=3)
        # _, fr = np.unique(bin_values, return_counts=True) #***
        vs = np.sum(np.stack((v, s)), axis=0)
        if self.verbose:
            print(' + Time to calculate features: {}\'s'.format(
                np.round(time.time() - start, 3)))
        # remove outliers
        start = time.time()
        data = pd.DataFrame({'val_dens': d, 'val_varskew': vs})
        out_keep = drop_outliers(
            data=data,
            outlier_func=self.outlier_func,
            drop_and=self.outlier_drop_and
        )
        if calculate_outlier2:
            self.out_keep_final = drop_outliers(
                data=pd.DataFrame({'val': d+vs}),
                outlier_func=self.outlier_func,
                drop_and=self.outlier_drop_and
            )
        # scale values
        ds = scale_values(data[out_keep])
        for v1 in ds.columns:
            data['{}_scaled'.format(v1)] = 0.0
            data['{}_scaled'.format(v1)][out_keep] = ds[v1]
        data['val_scaled'] = data[
            ['val_dens_scaled', 'val_varskew_scaled']].sum(axis=1)
        data['outlier_keep'] = out_keep
        # data['flowrate'] = fr
        if calculate_outlier2:
            data['outlier_keep_final'] = self.out_keep_final

        if self.verbose:
            print(' + Time to remove outliers + scale features: {}\'s'.format(
                np.round(time.time() - start, 3)))

        return data

    def apply_clean(self, data: pd.DataFrame):# -> pd.DataFrame:
        """Apply cleaning.

        Args:
            data (pandas.DataFrame): Object x feature matrix.

        Returns:
            pandas.DataFrame: See `MetaClean`.
        """
        out_keep = data['outlier_keep']

        # metaclean!
        data_ = data[out_keep]
        data_.index = range(len(data_))
        clean_keep, segments2, segments0 = self.metaclean.clean(
            data=data_,
            val_cols=['val_dens_scaled', 'val_varskew_scaled'],
            val_col_final='val_scaled'
        )
        data[self.fcs.bin_chan] = np.unique(self.fcs.data[self.fcs.bin_chan])
        data = data.assign(
            **{'clean_keep': False, 'segments_raw': -1, 'segments': -1})
        data['clean_keep'][out_keep] = clean_keep
        data['segments'][out_keep] = segments2
        data['segments_raw'][out_keep] = segments0

        return data

    def remove_outliers(
        self,
        data_merged: pd.DataFrame
    ):# -> pd.DataFrame:
        """
        After cleaning is done, revisits outliers to either keep all, some,
        or none of the outliers.

        Args:
            data_merged (pandas.DataFrame): object/row x feature matrix.

        Returns:
            pandas.DataFrame: `data_merged` where the `clean_keep` column is
                editted such that all, some, or no outliers are kept.
        """
        outs = data_merged['outlier_keep']
        if self.rm_outliers == 'some':
            outs = (outs & ~self.out_keep_final)
        if not any(outs) or self.rm_outliers == 'none':
            return None

        out_chpts = np.where( # not the case, but just in case
            np.diff(np.concatenate(([True], outs, [True]))) != 0)[0]

        if out_chpts[0] == 0:
            if len(out_chpts) == 2:
                return None

        out_chpts = out_chpts[2:]
        if out_chpts[-1] >= (len(outs) - 1):
            if len(out_chpts) == 2:
                return None

        out_chpts = out_chpts[:-2]
        clean_keepers = np.array(data_merged['clean_keep'])
        data_new = copy.copy(data_merged)
        for i in range(0, len(out_chpts), 2):
            keeper = all(clean_keepers[[out_chpts[i] - 1, out_chpts[i+1] + 1]])
            data_new['clean_keep'][out_chpts[i]:out_chpts[i+1]] = keeper

        return data_new



import numpy as np
import pandas as pd
from dataclasses import dataclass, InitVar

from metaclean3.channels import get_clean_time_chan, get_clean_fp_channels
from metaclean3.binning import get_time_bin_unit, get_time_binned

@dataclass(init=True)
class FCSfile:
    """Prepares an FCS file for MetaClean3.0.

    Attributes:
        data (pandas.DataFrame): FCS file data matrix sorted by time.
        time_step (float | None, optional):
            See :func:`binning.get_time_binned`. Defaults to None.
        time_chan (str): Time channel name. Defaults to 'time'.
        bin_chan (str, optional): Name of the bin column to be added.
        fluo_chans (list | np.ndarray, optional):
            Fluorescent channels that can be used for 0.
            Defaults to `np.array([])`.
        phys_chans (list | np.ndarray, optional):
            Physical morphological channels that can be used for 0.
            Defaults to `np.array([])`.
    """
    data: pd.DataFrame
    time_step: float | None = None # can only be None if time_bin is specified; meta.get('$TIMESTEP')
    time_chan: str = 'time'
    bin_chan: str = 'bin'
    fluo_chans: list | np.ndarray = np.array([])
    phys_chans: list | np.ndarray = np.array([])

    time_bin: InitVar[str | None] = None
    channel_unique_no: InitVar[int] = 25 # minimum number of unique values per column
    min_bin_size: InitVar[int] = 2000
    max_bin_size: InitVar[int] = 10000
    min_events_per_bin: InitVar[int] = 50

    def __post_init__(
        self,
        time_bin,
        channel_unique_no,
        min_bin_size,
        max_bin_size,
        min_events_per_bin
    ):
        """
        Args:
            time_bin (str, optional): See :func:`binning.get_time_binned`.
                Defaults to '1S'.
            channel_unique_no (int):
                See :func:`channels.get_clean_fp_channels`. Defaults to 25.
            min_bin_size (int, optional):
                See :func:`binning.get_time_binned`. Defaults to 2000.
            max_bin_size (int, optional):
                See :func:`binning.get_time_binned`. Defaults to 10000.
            min_events_per_bin (int, optional):
                See :func:`binning.get_time_binned`. Defaults to 50.
        """
        if len(self.data) == 0:
            raise ValueError('Data has no events.')

        # process channels: time; reorder data if needed
        self.data['index_original'] = self.data.index
        self.time_chan, self.data = get_clean_time_chan(
            data=self.data,
            time_chan=self.time_chan,
            min_bin_size=min_bin_size
        )
        # process channels: fluo, phys
        self.fluo_chans, self.phys_chans = get_clean_fp_channels(
            data=self.data,
            fluo_chans=self.fluo_chans,
            phys_chans=self.phys_chans,
            channel_unique_no=channel_unique_no
        )
        if self.fluo_chans is None or len(self.fluo_chans) == 0:
            raise ValueError('No fluorescent channels found.')

        # bin data by time
        time_bin = get_time_bin_unit(
            time_values=self.data[self.time_chan],
            time_bin=time_bin
        )
        self.data[self.bin_chan] = get_time_binned(
            time_values=self.data[self.time_chan],
            time_bin=time_bin,
            time_step=self.time_step,
            min_bin_size=min_bin_size,
            max_bin_size=max_bin_size,
            min_events_per_bin=min_events_per_bin
        )



import numpy as np
import pandas as pd
import ruptures as rpt

# TODO: shrinkable
def list_gains(
    chpts0M: list | np.ndarray | pd.Series,
    rpt_class: rpt.Binseg,
    chpts_inds: list | np.ndarray | pd.Series | None = None,
    dif: bool = True
):# -> np.ndarray:
    """List the gains for values fit on a ruptures class on given changepoints.

    Args:
        chpts0M (list | numpy.ndarray | pandas.Series): A vector listing the
            index of the first element in each segment + a last element that
            indicates the length of the fitted values.
        rpt_class (rpt.Binseg): A ruptures.Binseg object that is already fit on
            values we are calculating gains on.
        chpts_inds (list | numpy.ndarray | pandas.Series | None, optional):
            If the user only wants to calculate gains on a subset of the
            given changepoints, `chpts_inds` is a vector that lists the
            indices of `chpts_inds` for which a gain should be calculated.
            Defaults to None.
        dif (bool, optional): Whether to use gain difference (`True`) or to use
            gain ratio  (`False`). Defaults to True.

    Returns:
        numpy.ndarray: A 1D array containing a gain for each given changepoint.
    """
    gain = np.full((len(chpts0M)), -1.0)
    # prepare indices to calculate gain/gains for
    if chpts_inds is None:
        chpts_inds = np.array(range(1, len(chpts0M)-1))

    chpts_inds = np.array(chpts_inds)[
        np.isin(chpts_inds, np.array(range(1, len(chpts0M)-1)))]
    if len(chpts_inds) == 0:
        # warnings.warn('Chosen indices not in chpt list, returning -1s.')
        return gain

    chpts_inds = np.sort(chpts_inds)
    for ci in chpts_inds:
        start = chpts0M[ci-1]
        bkp = chpts0M[ci]
        end = chpts0M[ci+1]
        g = binseg_gain(start, bkp, end, rpt_class=rpt_class, dif=dif)
        gain[ci] = np.squeeze(np.array(g))

    return gain

def binseg_gain(
    start: int,
    bkp: int,
    end: int,
    rpt_class: rpt.Binseg,
    dif: bool = True
):# -> float:
    """Calculates the gain of a single changepoint.

    Args:
        start (int): Start index (first inde of first segment).
        bkp (int): Breakpoint index (first inde of second segment).
        end (int): End index (not inclusive).
        rpt_class (rpt.Binseg): A ruptures.Binseg (for example) object that is
            already fit on values we are calculating the gain on.
        dif (bool, optional): Whether to use gain difference (`True`) or to use
            gain ratio  (`False`). Defaults to True.

    Returns:
        float: Gain value.
    """
    segment_cost0 = rpt_class.cost.error(start, end)
    try:
        d = (rpt_class.cost.error(start, bkp) + rpt_class.cost.error(bkp, end))
        if dif:
            return segment_cost0 - d
        else:
            return segment_cost0 / d
    except rpt.exceptions.NotEnoughPoints:
        return 0

def chpts0M_to_segments(chpts0M: list | np.ndarray | pd.Series):# -> np.ndarray:
    """Converts changepoint indices to an integer segment label array.

    Args:
        chpts0M (list | numpy.ndarray | pandas.Series): Changepoint indices
            where the first value is 0 and the last value is the length of the
            desired segment label array.

    Returns:
        numpy.ndarray: A 1D integer segment label array.
    """
    if len(chpts0M) == 2:
        return np.full((chpts0M[-1]), 0)

    segments = np.full((chpts0M[-1]), -1)
    for i, ch in enumerate(chpts0M[:-1]):
        segments[ch:chpts0M[i+1]] = i
    return segments

def segments_to_chpts0M(segments: list | np.ndarray | pd.Series):# -> np.ndarray:
    """Converts an integer segment label array to changepoint indices.

    Args:
        segments (list | numpy.ndarray | pandas.Series):
            A 1D integer segment label array.

    Returns:
        numpy.ndarray: A 1D integer changepoint indices array.
    """
    chpts = np.where(np.diff(segments) != 0)[0] + 1
    chpts0M = np.concatenate(([0], chpts, [len(segments)]))
    return chpts0M




import pandas as pd
import numpy as np
from typing import Callable, Literal
from operator import gt, lt
import ruptures as rpt

from scipy.stats import gmean, ranksums, ttest_ind

from kneed import KneeLocator

from .utils import mode
from .changepoint import segments_to_chpts0M, list_gains, chpts0M_to_segments
# from .plot import plot_scat

# TODO longterm: shorten functions

# checks arguments chpts0M, r, and returns rl if it is none (using segments).
def _check_chpts0M_segments(
    chpts0M = None, segments = None, r = None, rl = None):# -> int | None:
    if not (chpts0M is None):
        if not (len(chpts0M) > 1 and chpts0M[0] == 0 and chpts0M[-1] > 0):
            raise ValueError('Invalid changepoints (chpts0M).')
        if not (r is None) and not (r >= 0 and r < chpts0M[-1]):
            raise ValueError('Invalid reference changepoint index (r).')
    if not (segments is None) and len(segments) > 0:
        if rl is None:
            return mode(segments)
        elif not (rl in segments):
            raise ValueError('Invalid reference segment label (rl).')
        else:
            return rl

def get_ref_ranges(
    ref_values: list,
    ignore_no: int = 5,
    percent_diff: float = 0.05,
    percent_shift: float = 0.1
):# -> tuple:
    """
    Get ranges (quantiles) for reference values.
    Intended for internal use; made public for external QA.

    Args:
        ref_values (list): A list of numeric vector(s) that the
            function will create ranges for.
        ignore_no (int, optional): See `in_percentiles()`. Defaults to 5.
        percent_diff (float, optional): See `in_percentiles()`. Defaults to 0.05.
        percent_shift (float, optional): See `in_percentiles()`. Defaults to 0.1.

    Returns:
        tuple:
            list: a list of four floats containing upper and lower bounds of
                maximum and minimum ranges.
            list: A list of floats representing the shift for each element in
                `ref_values`.
    """
    ref_range = [np.max(ref_values[0]), np.min(ref_values[0])] * 2
    ref_shift_all = []
    for ref_value in ref_values:
        if len(ref_value) < ignore_no:
            continue
        ref_percentiles = (
            np.quantile(ref_value, percent_diff),
            np.quantile(ref_value, 1-percent_diff))
        ref_shift = (percent_shift * (ref_percentiles[1] - ref_percentiles[0]))
        ref_shift_all.append(ref_shift)
        ref_range_ = [
            ref_percentiles[0] - ref_shift,
            ref_percentiles[0] + ref_shift,
            ref_percentiles[1] - ref_shift,
            ref_percentiles[1] + ref_shift]

        ref_range[0] = min(ref_range[0], ref_range_[0])
        ref_range[1] = max(ref_range[1], ref_range_[1])
        ref_range[2] = min(ref_range[2], ref_range_[2])
        ref_range[3] = max(ref_range[3], ref_range_[3])

    return ref_range, ref_shift_all

def get_test_ranges(
    ref_values: int,
    test_value: list | np.ndarray | pd.Series,
    small_no_per_seg: int = 50,
    percent_diff: float = 0.05,
    percent_shift: float = 0.1,
    lenient: bool = True
):# -> list:
    """
    Get ranges (quantiles) for test values.
    Intended for internal use only; made public for external QA.

    Args:
        ref_lens (int): The maximum length of value vectors in the
            reference segment; most of the time, this is the length of the
            reference segment, but an arbitrary length can be given.
        test_value (list | numpy.ndarray | pandas.Series): A numeric vector that
            the function will create ranges for.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.
        percent_diff (float, optional): See `in_percentiles()`. Defaults to 0.05.
        percent_shift (float, optional): See `in_percentiles()`. Defaults to 0.1.
        lenient (bool, optional): See `in_percentiles()`. Defaults to True.

    Returns:
        list: a list of four floats containing upper and lower bounds of
                maximum and minimum ranges.
    """
    if len(test_value) <= small_no_per_seg:
        test_range = [
            np.min(test_value),
            np.quantile(test_value, percent_diff), # * 2 to be 2x lenient
            np.quantile(test_value, 1-percent_diff),
            np.max(test_value)]
        # TODO: put into ref_func
        if lenient:
            ref_len = np.argmax([len(r) for r in ref_values])
            in_range = np.where(
                (test_range[0] <= ref_values[ref_len]) &
                (ref_values[ref_len] > test_range[3]))[0]
            in_m1 = np.diff(in_range) >= len(test_value) * (percent_diff * 2)
            if np.sum(in_m1) > 0:
                r0 = ref_values[ref_len]
                gi = np.where(in_m1)[0]
                ref_values.append(
                    np.concatenate([r0[in_range[g]:in_range[g+1]] for g in gi]))
    else:
        test_range = [
            np.quantile(test_value, percent_diff * (1 - percent_shift)),
            np.quantile(test_value, percent_diff * (1 + percent_shift)),
            np.quantile(test_value, 1 - (percent_diff * (1 + percent_shift))),
            np.quantile(test_value, 1 - (percent_diff * (1 - percent_shift)))]

    return test_range

def apply_ranges(
    ref_range: list | np.ndarray,
    test_range: list | np.ndarray
):# -> bool:
    """
    Test whether test ranges fall within reference ranges.
    Intended for internal use only; made public for external QA.

    Args:
        ref_range (list | numpy.ndarray): A reference ranges vector of length 4.
        test_range (list | numpy.ndarray): A test ranges vector of length 4.

    Returns:
        bool: Whether or not test ranges fall within reference ranges.
    """
    if not (len(test_range) == 4) or not (len(ref_range) == 4):
        raise ValueError('Ranges/quantiles must be of length 4.')
    cond = (
        ((test_range[0] >= ref_range[0] and ref_range[1] > test_range[0]) or
        (test_range[1] >= ref_range[0] and ref_range[1] > test_range[1]) or
        (test_range[0] < ref_range[0] and ref_range[1] <= test_range[1])) and
        ((test_range[2] >= ref_range[2] and ref_range[3] > test_range[2]) or
        (test_range[3] >= ref_range[2] and ref_range[3] > test_range[3]) or
        (test_range[2] <= ref_range[2] and ref_range[3] <= test_range[3]))
    )
    return cond

# calculate statistics from and merge/refine segments
def in_percentiles(
    ref_values: list,
    test_value: list | np.ndarray | pd.Series,
    check_mean: bool = False,
    check_var: bool = False,
    switch: bool = True,
    ignore_no: int = 5,
    small_no_per_seg: int = 50,
    percent_diff: float = 0.05,
    percent_shift: float = 0.1,
    lenient: bool = True
):# -> bool:
    """
    Test whether test and reference values have a similar distribution.
    Intended for internal use only; made public for external QA.

    Args:
        ref_values (list): A vector of reference values.
        test_value (list | numpy.ndarray | pandas.Series): A vector of test
            values.
        check_mean (bool, optional): Whether or not to check similarity of
            reference and test means. Defaults to False.
        check_var (bool, optional): Whether or not to check similarity of
            reference and test variance. Defaults to False.
        switch (bool, optional): Whether or not to check similarity of reference
            and test mean and variance if test ranges do not fall within
            reference ranges. Defaults to True.
        ignore_no (int, optional): Indicates the minimum length of an acceptable
            vector. Defaults to 5.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.
        percent_diff (float, optional): Quantile to be calculated from 0 and 1.
            Defaults to 0.05.
        percent_shift (float, optional): Percent above and below mean of values.
            Defaults to 0.1.
        lenient (bool, optional): If `test_value` is a small vector, users can
            choose whether or not the ranges should be lenient or not.
            Defaults to True.

    Returns:
        bool: Indicates whether or not test ranges are within reference ranges.
    """
    # if segment is small enough, just keep it
    rall = all([len(x) <= ignore_no for x in ref_values])
    if len(test_value) <= ignore_no or rall:
        return True

    # create reference segments ranges
    ref_range, ref_shift_all = get_ref_ranges(
        ref_values=ref_values, ignore_no=ignore_no,
        percent_diff=percent_diff, percent_shift=percent_shift
    )
    # create test segment ranges + adjust ref values
    test_range = get_test_ranges(
        ref_values=ref_values, test_value=test_value,
        small_no_per_seg=small_no_per_seg,
        percent_diff=percent_diff, percent_shift=percent_shift, lenient=lenient
    )
    # compare test and reference segment ranges
    cond = apply_ranges(ref_range, test_range)
    # compare test and reference segment variances
    if check_var:
        cond_ = False
        tvar = (np.quantile(test_value, 1-percent_diff) -
                np.quantile(test_value, percent_diff)) # np.var(test_value)
        for ref_value, ref_shift in zip(ref_values, ref_shift_all):
            rvar = (np.quantile(ref_value, 1-percent_diff) -
                    np.quantile(ref_value, percent_diff)) # np.var(test_value)
            cond_ = cond_ or (np.abs(tvar-rvar) < ref_shift)
        cond = cond and cond_
    # compare test and reference segment means
    if check_mean:
        cond_ = False
        neg_diff = min(np.min(test_value), 0)
        tmean = gmean(test_value - neg_diff, nan_policy='omit') + neg_diff
        for ref_value, ref_shift in zip(ref_values, ref_shift_all):
            neg_diff = min(np.min(ref_value), 0)
            rmean = gmean(ref_value - neg_diff, nan_policy='omit') + neg_diff
            cond_ = cond_ or (
                tmean >= rmean - ref_shift and tmean < rmean + ref_shift)
        cond = cond and cond_
    # compare test and reference segment var/means only if range test fails
    if switch and not cond:
        cond = False
        tvar = (np.quantile(test_value, 1-percent_diff) -
                np.quantile(test_value, percent_diff))
        for ref_value, ref_shift in zip(ref_values, ref_shift_all):
            rvar = (np.quantile(ref_value, 1-percent_diff) -
                    np.quantile(ref_value, percent_diff)) # np.var(test_value)
            neg_diff = min(np.min(np.concatenate((test_value, ref_value))), 0)
            tmean = gmean(test_value - neg_diff, nan_policy='omit')
            rmean = gmean(ref_value - neg_diff, nan_policy='omit')
            cond_ = (np.abs(tvar - rvar) < ref_shift) and (
                     tmean >= rmean - ref_shift and tmean < rmean + ref_shift)
            cond = cond or cond_

    return cond

def get_ref_updown(
    r: int,
    chpts0M: list | np.ndarray | pd.Series,
    vlist: list,
    ref_percents: list = [1, 0.5],
    ref_percent: float = 0.4
):# -> list:
    """
    Extracts values for reference segment.
    Intended for internal use only; made public for external QA.

    Args:
        r (int): Index of the reference segment in `chpts0M`.
        chpts0M (list | numpy.ndarray | pandas.Series): A vector containing the
            first index of each segment.
        vlist (list): List of numeric vector of values.
        ref_percents (list, optional): List of percent of reference values
            next to the test values to extract. Defaults to [1, 0.5].
        ref_percent (float, optional): Percent of length of `vlist`;
            a minimum of this value and the length of reference values is taken
            as another alternative to how many reference next to the test values
            to extract. Defaults to 0.4.

    Returns:
        list: List of reference values extracted.
    """
    _check_chpts0M_segments(chpts0M=chpts0M, r=r)
    rps = np.append(
        [int((chpts0M[r+1]-chpts0M[r]) * p) for p in ref_percents],
        [min(int(chpts0M[-1] * ref_percent), chpts0M[r+1] - chpts0M[r])])
    ref_values = [
        [[np.array(v[chpts0M[r]:(chpts0M[r]+rp)])
            for rp in rps] for v in vlist],
        [[np.array(v[(chpts0M[r+1]-rp):chpts0M[r+1]])
            for rp in rps] for v in vlist]]
    return ref_values

def go_updown(
    r: int,
    chpts0M: list | np.ndarray,
    lim: int,
    refs: list,
    vlist: list,
    ip_func: Callable = in_percentiles,
    exception_no: int = 2,
    ignore_no: int = 5
):# -> tuple:
    """
    Tests whether reference segment is similar to its adjacent segment(s).
    Intended for internal use only; made public for external QA.

    Args:
        r (int): Index of the reference segment in `chpts0M`.
        chpts0M (list | numpy.ndarray | pandas.Series): A vector containing the
            first index of each segment.
        lim (int): Upper limit for the index of segments to test i.e. the index
            of the end or the next reference segment to test .
        refs (list): List of reference vectors.
        vlist (list): List of all value vectors.
        ip_func (Callable, optional): Function used to determine whether test
            and reference values are similar. Defaults to `in_percentiles()`.
        exception_no (int, optional): Users can choose to check test values
            `exception_no` segments away from reference segment
            (i.e. non-adjacent test segments). Defaults to 2.
        ignore_no (int, optional): See `in_percentiles()`. Defaults to 5.

    Returns:
        tuple:
            bool: Whether test values and adjacent reference values have similar
                distribution.
            numpy.ndarray: Indices of segments that were skipped in between test
                and similar reference segment.
    """
    # set parameters
    _check_chpts0M_segments(chpts0M=chpts0M, r=r)
    go_up = r < lim
    op = lt if go_up else gt
    ec = 1 if go_up else -1
    df = ec * 2

    # start: check segments before and after segment r for merging
    en = 0
    go = False
    skipped_chpt_ = np.array([], dtype=int)
    while not go and op(r+ec+en, lim) and abs(en) < exception_no:
        while ((en > 0) and op(r+df+en, lim) and (
            chpts0M[r+ec+1+en]-chpts0M[r+ec+en] <= ignore_no)):
            s = chpts0M[r+ec+en]
            if s not in skipped_chpt_:
                skipped_chpt_ = np.concatenate((
                    skipped_chpt_, [s, chpts0M[r+ec+1+en]]))
            en += ec
        if op((r+ec+en), lim):
            go = True
            tests = [v[chpts0M[r+ec+en]:chpts0M[r+ec+1+en]] for v in vlist]
            for test_value, ref_values in zip(tests, refs):
                rli = np.argmax([len(rl) for rl in ref_values])
                new_len = min(len(ref_values[rli]), len(test_value))
                if len(ref_values[rli]) > new_len:
                    if go_up:
                        ref_values.append(ref_values[rli][-new_len:])
                    else:
                        ref_values.append(ref_values[rli][:new_len])
                go_ = ip_func(ref_values=ref_values, test_value=test_value,
                              check_mean=en != 0)
                go = go and go_
            s = chpts0M[r+ec+en]
            if not go and s not in skipped_chpt_:
                skipped_chpt_ = np.concatenate((
                    skipped_chpt_, [s, chpts0M[r+ec+1+en]]))
        if not go:
            en += ec
    return go, skipped_chpt_

def get_ref_func2(
    vlist: list,
    ref_inds: np.ndarray,
    test_i1: int,
    test_i2: int,
    # test_i1_raw: int | None = None,
    # test_i2_raw: int | None = None,
    small_no_per_seg: int = 50,
    small_ref_prop: float = 0.5,
    sorted_index: np.ndarray | None = None
):# -> list:
    """
    Extracts values for reference segment; this function is an alternative to
    `get_ref_updown()` for when the user wants to subset values in
    the reference segment that are closest to some test segment. The reference
    segment indices do not need to be continuous.
    Intended for internal use only; made public for external QA.

    Args:
        vlist (list): List of value vectors.
        ref_inds (numpy.ndarray): Indices of reference values.
        test_i1 (int): Starting index of test segment.
        test_i2 (int): Ending index of test segment.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.
        small_ref_prop (float, optional): Proportion of reference segment to
            test. Defaults to 0.5.
        sorted_index (numpy.ndarray | None, optional): Indices sorted based on
            segment quantiles; see `utils.order_value_bins()`. Defaults to None.

    Returns:
        list: List of reference value vectors.
    """
    # test_i1_raw = test_i1_raw if not (test_i1_raw is None) else test_i1
    # test_i2_raw = test_i2_raw if not (test_i2_raw is None) else test_i2
    test_len = test_i2 - test_i1
    big_test = test_len > small_no_per_seg and len(ref_inds) > test_len
    refs = [[v[ref_inds]] for v in vlist]
    for vi in range(len(vlist)):
        v = vlist[vi]
        if big_test:
            med_diff = ref_inds - np.median([test_i1, test_i2-1])
            ref_close_inds = np.argsort(np.abs(med_diff))[:test_len]
            refs[vi].append(v[ref_inds[ref_close_inds]])
            if not (sorted_index is None):
                sorted_ref_inds = np.where(np.isin(sorted_index, ref_inds))[0]
                med_diff = sorted_ref_inds - np.median(
                    np.where((sorted_index==test_i1) |
                             (sorted_index==(test_i2-1)))[0])
                goodi = sorted_ref_inds[np.argsort(np.abs(med_diff))[:test_len]]
                ref_close_inds = sorted_index[goodi]
                refs[vi].append(v[ref_close_inds])
        elif test_len <= small_no_per_seg:
            smin = np.min(vlist[vi][test_i1:test_i2])
            smax = np.max(vlist[vi][test_i1:test_i2])
            in_r = np.append([0], np.array(
                (smin <= v[ref_inds]) & (v[ref_inds] <= smax), dtype=int))
            startj = np.where(np.diff(in_r) == 1)[0]
            endj = np.where(np.diff(in_r) == -1)[0]
            lenj = min(len(startj), len(endj))
            sei = np.where(endj[:lenj] - startj[:lenj] >=
                           test_len * small_ref_prop)[0]
            if len(sei) > 0:
                sei_ref = np.concatenate(
                    [np.array(v[ref_inds[startj[j]:endj[j]]]) for j in sei])
                refs[vi].append(sei_ref)
    return refs

def merge_segments_test(
    values: list | np.ndarray | pd.Series,
    segments: list | np.ndarray | pd.Series,
    merge_signif_test: Literal['t', 'wilcox'] = 'wilcox',
    p_thres: float = 0.05,
    ignore_no: int = 5,
    strict: bool = True
):# -> np.ndarray:
    """
    Merge segments that contain values that are not significantly different
    based on some significance test.
    Intended for internal use only; made public for external QA.

    Args:
        values (list | numpy.ndarray | pandas.Series): Vector of values.
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        merge_signif_test (str, optional): Type of significance test to used
            i.e. 't' (T-test) or 'wilcox' (Wilcoxan). Defaults to 'wilcox'.
        p_thres (float, optional): P-value threshold. Defaults to 0.05.
        ignore_no (int, optional): See `in_percentiles()`. Defaults to 5.
        strict (bool, optional): If set to `True`, function will only test
            adjacent segments; otherwise, . Defaults to True.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) < 2:
        return segments

    chpts0M = segments_to_chpts0M(segments)

    # determine which changepoints to keep based on significance test
    del_inds = [0]
    w_i = 0
    while len(del_inds) > 0 and len(chpts0M) > 0 and not (strict and w_i > 0):
        del_inds = []
        ps = []
        lengthi = np.diff(chpts0M)
        for i in range(1, len(chpts0M) - 1):
            x = np.array(values[chpts0M[i-1]:chpts0M[i]])
            x = x[~np.isnan(x)]
            y = np.array(values[chpts0M[i]:chpts0M[i+1]])
            y = y[~np.isnan(y)]
            if merge_signif_test == 'wilcox':
                p = ranksums(x, y).pvalue
            else:
                p = ttest_ind(x, y).pvalue
            # sometimes, small segments could be falsely p >= p_thres
            # so we check the segments beside it; even if the next segment is
            # also small, there's less chance of false positive?
            if (lengthi[i-1] <= ignore_no and len(y) > ignore_no and i > 1 and
                chpts0M[i-1] - chpts0M[i-2] > ignore_no and
                ps[-1] > p_thres and p > p_thres):
                x = np.array(values[chpts0M[i-2]:chpts0M[i-1]])
                x = x[~np.isnan(x)]
                if merge_signif_test == 'wilcox':
                    p_ = ranksums(x, y).pvalue
                else:
                    p_ = ttest_ind(x, y).pvalue
                if p_ < p_thres:
                    if ps[-1] < p:
                        ps[-1] = p_
                    else:
                        p = p_
            ps.append(p)
        w_i += 1
        del_inds = np.where(np.array(ps) >= p_thres)[0]
        if len(del_inds) > 0:
            chpts0M = np.delete(chpts0M, del_inds + 1)

    return chpts0M_to_segments(chpts0M)

def merge_segments_gain(
    values: list | np.ndarray | pd.Series,
    segments: list | np.ndarray,
    cost_model: str = 'rank',
    binseg_jump: int = 2,
    binseg_min_size: int = 2,
    knee: bool = False,
    min_ref_percent: float = 0.4
    # png_dir: str = ''
):# -> np.ndarray:
    """
    Merge segments with minimum gain at their changepoints until one segment is
    at least `min_ref_percent` percent of the maximum length.
    Intended for internal use only; made public for external QA.

    Args:
        values (list | numpy.ndarray | pandas.Series): Vector of values.
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        cost_model (str, optional): Type of cost calculation method to use;
            see `ruptures.Binseg`. Defaults to 'rank'.
        binseg_jump (int, optional): See argument `jump` in `ruptures.Binseg`.
            Defaults to 2.
        binseg_min_size (int, optional): See argument `min_size` in
            `ruptures.Binseg`. Defaults to 2.
        knee (bool, optional): Use gain value knee as stopping criteria. For
            testing purposes only, do not change. Defaults to False.
        min_ref_percent (float, optional): Stopping criteria; indicates the
            minimum length (proportion of total) of any segment.
            Defaults to 0.4.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging segments.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) == 1:
        return segments

    chpts0M = segments_to_chpts0M(segments)
    if len(chpts0M) < (1 + (1 / min_ref_percent)):
        return segments

    # merge chpts until a segment is at least min_ref_percent long
    rpt_class = rpt.Binseg(
        model=cost_model, jump=binseg_jump, min_size=binseg_min_size)
    rpt_class.fit(signal=np.array(values))
    gains = list_gains(rpt_class=rpt_class, chpts0M=chpts0M)
    thres_met = any(np.diff(chpts0M) / len(values) > min_ref_percent)
    # use gain value knee (i.e. elbow) as the merge limit if specified.
    if knee: # TODO 10
        kneedle = KneeLocator(
            np.array(range(2, len(gains))),
            np.sort(gains[1:-1]),
            S=1.0, curve="convex", direction="increasing")
        if not (kneedle.elbow is None) and not (kneedle.knee is None):
            keep_ind = np.argsort(gains)[min(kneedle.elbow, kneedle.knee):]
            if len(keep_ind) > 0:
                keep_ind = np.sort(keep_ind)
                chpts0M = np.concatenate(
                    ([0], chpts0M[keep_ind], [chpts0M[-1]]))
        else:
            knee = False
    if not knee: # TODO: mean variance must be within something...
        thres_met = any(np.diff(chpts0M) / len(values) > min_ref_percent)
        while len(chpts0M) > 2 and not thres_met:
            # delete chpt with lowest gain
            del_ind = np.argmin(gains[1:-1]) + 1
            chpts0M = np.delete(chpts0M, del_ind)
            gains = np.delete(gains, del_ind)
            thres_met = any(np.diff(chpts0M) / len(values) > min_ref_percent)

            # recalculate gain
            if len(chpts0M) > 2:
                gains_new = list_gains(
                    chpts0M=chpts0M, chpts_inds=[del_ind-1, del_ind],
                    rpt_class=rpt_class)
                gains[gains_new != -1] = gains_new[gains_new != -1]

    segments = chpts0M_to_segments(chpts0M)
    return np.array(segments)

def merge_segments_quant(
    vlist: list,
    segments: list | np.ndarray | pd.Series,
    ip_func: Callable = in_percentiles,
    get_ref_fun: Callable = get_ref_updown,
    small_no_per_seg: int = 50,
    keep_skipped: str = 'all'
    # png_dir: str = ''
):# -> np.ndarray:
    """
    Merge segments if their quantiles are within a range of each other.
    Intended for internal use only; made public for external QA.

    Args:
        vlist (list): List of value vectors.
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        ip_func (Callable, optional): Function used to determine whether test
            and reference values are similar. Defaults to `in_percentiles()`.
        get_ref_fun (Callable, optional): A function that extracts a list of
            reference values. Defaults to `get_ref_updown()`.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.
        keep_skipped (str, optional): Indicates whether to keep all, some, or no
            skipped segments i.e. 'all', 'some', 'none'. Defaults to 'all'.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging segments.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) == 1:
        return segments

    # prepare arguments
    chpts0M = segments_to_chpts0M(segments=segments)
    if len(chpts0M) < 3:
        return segments

    # check adjacent segments against longest (reference) segments
    max_list = np.array([], dtype=int)
    skipped = np.array([], dtype=int)
    # plot_i = 0
    primary_twice = False
    while ((len(chpts0M) > 0) and
           (not all(np.isin(chpts0M[:-1], max_list)) or not primary_twice)):
        r_list = np.argsort(np.diff(chpts0M))
        in_max = np.isin(r_list, np.where(np.isin(chpts0M[:-1], max_list)))
        r = mode(chpts0M_to_segments(chpts0M))
        if all(in_max):
            if primary_twice:
                break
            else:
                primary_twice = True
        # check if segment is large enough to become a reference segment
        elif not primary_twice:
            r = r_list[~in_max][-1]
        if chpts0M[r+1] - chpts0M[r] <= small_no_per_seg:
            max_list = np.append(max_list, chpts0M[r])
            continue
        # set upper/lower bounds for reference segments: half, %, all
        ref_values = get_ref_fun(r=r, chpts0M=chpts0M, vlist=vlist)
        # set up/lower limits for segments to check
        lims = [-1, len(chpts0M) - 1]
        if len(max_list) > 0:
            up_max_ = np.where(np.isin(chpts0M[r:-1], max_list))[0]
            down_min_ = np.where(np.isin(chpts0M[:r], max_list))[-1]
            if len(up_max_) > 0:
                lims[1] = r + up_max_[0]
            if len(down_min_) > 0:
                lims[0] = down_min_[0]
        gos = [r - 1 > lims[0], r + 1 < lims[1]]
        # plot_j = 0
        # merge warrented segments
        while any(gos):
            for i in [1, 0]:
                if gos[i]:
                    # explicitly assign variables for testing
                    gos[i], skipped_ = go_updown(
                        r=r, chpts0M=chpts0M,
                        lim=lims[i], refs=ref_values[i],
                        vlist=vlist, ip_func=ip_func)
                if gos[i]:
                    chpts0M = np.delete(chpts0M, r+1 if i == 1 else r)
                    r = r - 1 if i == 0 else r
                    skipped = np.append(skipped, skipped_)
            if np.any(gos):
                ref_values = get_ref_fun(r=r, chpts0M=chpts0M, vlist=vlist)
                lims[1] -= np.sum(gos)
                gos = [r - 1 > lims[0] and gos[0], r + 1 < lims[1] and gos[1]]

            # # plot combined segments
            # if png_dir:
            #     plot_scat(
            #         np.array(range(len(vlist[0]))),
            #         np.sum(np.stack(vlist), axis=0),
            #         L=chpts0M_to_segments(chpts0M),
            #         plot_title='percentile combined segments SORTED',
            #         out_path='{}/2.while_{}.{}_SORTED.png'.format(
            #             png_dir, plot_i, plot_j))
            # plot_j += 1

        # plot_i += 1
        max_list = np.append(max_list, chpts0M[r])

    # keep or removed skipped segments
    r = np.argmax(np.diff(chpts0M))
    segments = chpts0M_to_segments(chpts0M)
    skipped_chpt = np.array(np.unique(skipped), dtype=int)
    skipped_chpt = skipped_chpt[
        (skipped_chpt >= chpts0M[r]) & (skipped_chpt < chpts0M[r+1])]
    if keep_skipped != 'all':
        for i, ch in enumerate(skipped_chpt[0::2]):
            segments[ch:skipped_chpt[i+1]] = -1

    return segments

def refine_reference_segment(
    vlist: list,
    segments: list | np.ndarray,
    rl: int | None = None,
    sorted_index: np.ndarray | None = None,
    ip_func: Callable = in_percentiles,
    small_no_per_seg: int = 50,
    overlap_no: int = 0,
    split_by_segment: bool = True,
    multi_refine: bool = True
):# -> np.ndarray:
    """
    Check if there are sporadic segments within the reference segments that
    should be merged with the reference segment.
    Intended for internal use only; made public for external QA.

    Args:
        vlist (list): List of value vectors.
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        rl (int | None, optional): Reference segment label; if set to `None`,
            sets the reference segment as the largest segment. Defaults to None.
        sorted_index (numpy.ndarray | None, optional): Indices sorted based on
            segment quantiles; see `utils.order_value_bins()`. Defaults to None.
        ip_func (Callable, optional): Function used to determine whether test
            and reference values are similar. Defaults to `in_percentiles()`.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.
        overlap_no (int, optional): Numer of values allowed to overlap.
            Defaults to 0.
        split_by_segment (bool, optional): If there are multiple segments
            in between the reference segment, choose whether to test them
            separately. Defaults to True.
        multi_refine (bool, optional): Optional, if set to True, function will
            refine repeatedly. Defaults to True.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging segments.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) == 1:
        return segments

    # start: merge segments into reference if similar enough
    rl = _check_chpts0M_segments(segments=segments, rl=rl)
    any_removed = True
    while any_removed:
        # get indices of segments in between reference segment
        ri = np.where(segments == rl)[0]
        chpts0M = segments_to_chpts0M(segments)
        ri_diff = np.diff(ri)
        starti = ri[np.where(np.append(ri_diff, [0]) > 1)[0]] + 1
        if len(starti) == 0 and ri[0] == 0 and ri[-1] == (len(segments)-1):
            break
        lengthi = ri_diff[ri_diff > 1]
        endi = starti + lengthi - 1
        if ri[0] > 0:
            endi = np.append(ri[0], endi)
            start_chpts0M = chpts0M[np.where(chpts0M == ri[0])[0] - 1]
            starti = np.append(start_chpts0M, starti)
        if ri[-1] < (len(segments)-1):
            starti = np.append(starti, ri[-1] + 1)
            end_chpts0M = chpts0M[np.where(chpts0M == ri[-1] + 1)[0] + 1]
            endi = np.append(endi, end_chpts0M)

        # for each segment inside the reference segment...
        any_removed = False
        for i in range(len(starti)):
            chs = np.unique(segments[starti[i]:endi[i]])
            to_keep = len(chs) == 1 or endi[i] - starti[i] <= small_no_per_seg
            # there is only one segment or if the inbetween segment is small,
            # test normally.
            if to_keep:
                refs = get_ref_func2(
                    vlist=vlist, ref_inds=ri,
                    test_i1=starti[i], test_i2=endi[i],
                    small_no_per_seg=small_no_per_seg,
                    sorted_index=sorted_index)
                starti_ = max(0, starti[i] - overlap_no)
                endi_ = min(len(segments), endi[i] + overlap_no)
                tests = [v[starti_:endi_] for v in vlist]
                for test_value, ref_values in zip(tests, refs):
                    to_keep_ = ip_func(
                        ref_values=ref_values, test_value=test_value,
                        check_mean=endi[i] == ri[0] or ri[-1] < starti[i])
                    to_keep = to_keep and to_keep_
                if to_keep:
                    segments[starti[i]:endi[i]] = rl
            elif len(chs) > 1:
                # test each segment inbetween or combine them for testing
                if split_by_segment:
                    ch_inds = [starti[i] + np.where(
                        segments[starti[i]:endi[i]] == c)[0] for c in chs]
                else:
                    chs = np.where(np.diff(segments[starti[i]:endi[i]])!=0)[0]+1
                    chs = np.concatenate(([0], chs, [endi[i] - starti[i]]))
                    ch_inds = [starti[i] + np.array(range(chs[ci], chs[ci+1]))
                                for ci in range(len(chs) - 1)]
                for ci in ch_inds:
                    if ci[0] > 0 and overlap_no > 0:
                        ci = np.concatenate((np.array(range(
                            ci[0] - overlap_no, ci[0]), dtype=int), ci))
                    if ci[-1] < len(segments) and overlap_no > 0:
                        ci = np.concatenate((ci, np.array(range(
                            ci[-1] + 1, ci[-1] + overlap_no + 1), dtype=int)))
                    # TODO: come back to this, maybe not change ref inds?
                    refs = get_ref_func2(
                        vlist=vlist, ref_inds=ri,
                        test_i1=ci[0], test_i2=ci[-1],
                        small_no_per_seg=small_no_per_seg,
                        sorted_index=sorted_index)
                    tests = [v[ci] for v in vlist]
                    to_keep = True
                    for test_value, ref_values in zip(tests, refs):
                        to_keep_ = ip_func(
                            ref_values=ref_values, test_value=test_value,
                            lenient=False)
                        to_keep = to_keep and to_keep_
                    if to_keep:
                        segments[ci] = rl

            if not any_removed and to_keep and multi_refine:
                any_removed = True

    return np.array(segments)

def refine_reference_segment_inbetween(
    segments: list | np.ndarray,
    rl: int | None = None,
    small_no_per_seg: int = 50
):# -> np.ndarray:
    """
    Same as `refine_reference_segment()` but intended for when there are
    sporadic parts of the reference segment between other segments.
    Intended for internal use only; made public for external QA.

    Args:
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        rl (int | None, optional): Reference segment label; if set to `None`,
            sets the reference segment as the largest segment. Defaults to None.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging segments.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) == 1:
        return segments

    rl = _check_chpts0M_segments(segments=segments, rl=rl)

    # get indices of reference segments in between other segments
    chpts0M = segments_to_chpts0M(segments)
    ref_inds = np.where(segments == rl)[0]
    ref_inds = np.concatenate(([-1], ref_inds, [len(segments)]))
    ri_diff = np.diff(ref_inds)
    endi = ref_inds[np.where(np.append(ri_diff, [0]) > 1)[0]] + 1
    if len(endi) == 0:
        return np.array(segments)
    starti = chpts0M[np.where(np.isin(chpts0M, endi))[0] - 1]
    lengthi = endi - starti
    chi = np.where(lengthi <= small_no_per_seg)[0]
    if len(chi) == 0:
        return segments

    # TODO: temporary
    for i in chi:
        around_ref = np.concatenate((
            segments[max(0, starti[i] - small_no_per_seg):starti[i]],
            segments[endi[i]:min(chpts0M[-1], endi[i] + small_no_per_seg)]))
        if all(around_ref != rl):
            segments[starti[i]:endi[i]] = -1

    return np.array(segments)

def remove_reference_ends(
    values: list | np.ndarray | pd.Series,
    segments: list | np.ndarray | pd.Series,
    rl: int | None = None,
    ip_func: Callable = in_percentiles,
    small_no_per_seg: int = 50
):# -> np.ndarray:
    """
    Check ends of the reference segment to see if there are any sporacidity.
    Merge ends if applicable.
    Intended for internal use only; made public for external QA.

    Args:
        values (list | numpy.ndarray | pandas.Series): Vector of values.
        segments (list | numpy.ndarray | pandas.Series): Segment label vector.
        rl (int | None, optional): Reference segment label; if set to `None`,
            sets the reference segment as the largest segment. Defaults to None.
        ip_func (Callable, optional): Function used to determine whether test
            and reference values are similar. Defaults to `in_percentiles()`.
        small_no_per_seg (int, optional): Vectors with a smaller length than
            this value is considered a "small" vector;
            small vectors are treated differently because
            smaller segments can contain more sporadic values. Defaults to 50.

    Returns:
        numpy.ndarray: an updated version of `segments` after merging segments.
    """
    segments = np.array(segments)
    if len(np.unique(segments)) == 1:
        return np.array(segments)

    rl = _check_chpts0M_segments(segments=segments, rl=rl)
    ri = segments == rl
    ref_inds = np.where(ri)[0]
    ri_diff = np.diff(ref_inds)
    starti = ref_inds[np.where(np.append(ri_diff, [0]) > 1)[0]] + 1
    if len(starti) == 0:
        return segments
    lengthi = ri_diff[ri_diff > 1]
    endi = starti + lengthi - 1
    # check if there are inserts into the reference segment at ends
    cond_ = [starti[0] <= small_no_per_seg,
                endi[-1] >= (len(values) - small_no_per_seg - 1)]
    refs_ = [values[starti[0]:endi[0]], values[starti[-1]:endi[-1]]]
    tests_ = [values[np.where(ri[:starti[0]])[0]],
                values[endi[-1] + np.where(ri[endi[-1]:])[0]]]
    inds_ = [np.array(range(starti[0])),
                np.array(range(endi[-1], len(segments)))]
    for cond, refs, tests, inds in zip(cond_, refs_, tests_, inds_):
        if cond:
            rm_values = ip_func(ref_values=[refs], test_value=tests)
            if rm_values:
                segments[inds] = -1

    return segments







import numpy as np
import copy
import warnings
import re
import difflib
import pandas as pd
from scipy.stats import gmean

## generic functions ####
def get_close_match(
    word: str,
    list_of_words: list
):# -> str:
    """Returns string from list_of_words that is the closest match to word.

    Args:
        word (str): A string.
        list_of_words (list[str]): A list of strings word will match to.

    Returns:
        str: string from list_of_words that is the closest match to word.
    """
    if len(list_of_words) == 1:
        return list_of_words[0]

    max_ind = 0
    max_score = difflib.SequenceMatcher(None, word, list_of_words[0]).ratio()
    for i in range(1, len(list_of_words)):
        word_ = list_of_words[i]
        score = difflib.SequenceMatcher(
            None, word.lower(), word_.lower()).ratio()
        if score > max_score:
            max_score = score
            max_ind = i

    return list_of_words[max_ind]

def align_list(
    list1: list,
    list2: list
):# -> list:
    """Converts list2 into list1.

    We assume both lists should be the same and we are only cleaning up list2.
    If some element in list2 does not exist in list1, this function finds
    the most similar string in list1 and replaces that element.

    Args:
        list1 (list[str]): List to compare to.
        list2 (list[str]): List to convert into list2.

    Returns:
        list2_ (list[str]): Cleaned-up version of list2.
    """

    list2_ = list2.copy()
    if len(list2_) > 0:
        for mi in range(len(list2_)):
            if list2_[mi] not in list1:
                list2_[mi] = get_close_match(list2_[mi], list1)
        list2_l = set(list2_)
        if len(list2_l) < len(list2):
            msg = 'Elements in list2 not found in list1; they will be removed.'
            warnings.warn(msg)
            list2_ = list(list2_l)

    return list2_

def arg_lim(
    x: int | float,
    xmin: int | float = -np.Inf,
    xmax: int | float = np.Inf
):
    """Sets the lower and upper limit of a given numeric variable.

    Args:
        x (int | float): Numeric variable.
        xmin (int | float, optional): Lower limit. Defaults to -numpy.Inf.
        xmax (int | float, optional): Upper limit. Defaults to numpy.Inf.

    Returns:
        int | float: `x` within the given lower and upper limit.
    """
    if xmin > xmax:
        warnings.warn('`xmin` can\'t be larger than `xmax`, ignoring limits.')
        return x

    return type(x)(max(xmin, min(xmax, x)))

def is_monotonic(x: np.ndarray, strict: bool = True):# -> bool:
    """Determines whether a vector is increasing or decreasing monotonically.

    Args:
        x: Any integer/float vector that can be converted to a 1D numpy array.
        strict (bool, optional): Whether or not to test strict monotonicity.
            Defaults to True.

    Returns:
        bool: Given vector is increasing or decreasing monotonically.
    """
    if len(x) < 2:
        return False

    x = np.array(x)
    if strict:
        increasing = all(y < z for y, z in zip(x, x[1:]))
        decreasing = all(y > z for y, z in zip(x, x[1:]))
    else:
        increasing = all(y <= z for y, z in zip(x, x[1:]))
        decreasing = all(y >= z for y, z in zip(x, x[1:]))

    res = increasing or decreasing
    return res

def mode(x: int | float):# -> int | float:
    """Returns the mode in vector `x`.

    Args:
        x: Any integer/float vector that can be converted to a 1D numpy array.

    Returns:
        int | float: The mode in vector `x`.
    """
    xu, xc = np.unique(x, return_counts=True)
    return xu[np.argmax(xc)]

def str_to_time(time_bin: str):# -> tuple:
    """Converts string to time format.

    Args:
        time_bin (str): Time in string format.

    Returns:
        tuple:
            float: Time value.
            str: Time unit.
    """
    tbs = re.findall(r'\d+', time_bin)
    time_scale = re.sub(r'\d+', '', time_bin)
    time_scale = time_scale.replace('.', '')
    tb = 1.0
    if len(tbs) > 0:
        if len(tbs) == 1:
            tb = float(tbs[0])
        else:
            tb = float(tbs[0]+'.'+''.join(tbs[1:]))

    return tb, time_scale


## remove duplicates in 2D arrays
def randomize_duplicates(
    arr: np.ndarray,
    noise_prop: float = 1 / 50000,
    strict: bool = False,
    lenient_percent: float = 0.99,
    seed: int = 123
):# -> np.ndarray: # 2D only, for rows
    """If there are any duplicate rows, adds random noise to last column.

    Args:
        arr (numpy.ndarray): 2D array
        noise_prop (float, optional): Proportion of noise.
            Defaults to 1/10000.
        strict (bool, optional): Actually find duplicates? This takes
            too long, so if false, we just find duplicate of the row sums.
            Defaults to False.duplicated_rows
        lenient_percent (float): If this percent of rows are unique,
            the function returns the original data. Defaults to 0.99.
        seed (int): Random seed. Defaults to 123.

    Returns:
        numpy.ndarray: 2D array with (almost, if not `strict`) unique rows.
    """
    arr = copy.copy(np.array(arr).astype(np.float32))

    # check if there are duplicates
    duplicated_inds, duplicated_size = duplicated_rows(
        arr, strict=strict, lenient_percent=lenient_percent)
    if duplicated_size < len(arr) * (1 - lenient_percent):
        return arr

    wm = 'There are {}/{} duplicate events ({}), will perturb values.'.format(
        duplicated_size, len(arr), 'rows' if strict else 'row sums')
    warnings.warn(wm)

    # perturb values
    ncol = arr.shape[1]
    p25 = np.quantile(arr[:, 0], .25)
    p75 = np.quantile(arr[:, 0], .75)
    np.random.seed(seed)
    noise = np.random.normal(
        loc=0.0, scale=noise_prop * (p75-p25), size=duplicated_size)
    arr[duplicated_inds, (ncol-1)] = arr[
        duplicated_inds, (ncol-1)] + noise

    return arr

def duplicated_rows(
    arr: np.ndarray,
    strict: bool = False,
    lenient_percent: float = 1.0
):# -> tuple:
    """Identifies duplicate rows in the given 2D array.

    Args:
        arr (numpy.ndarray): A 2D numpy array.
        strict (bool, optional): If set to True, finds rows with duplicate
            value. If set to False, finds rows whose sum is duplicated.
            Defaults to False.
        lenient_percent (float): If this percent of rows are unique,
            the function returns the original data. Defaults to 1.0.

    Returns:
        tuple:
            numpy.ndarray: 1D bool numpy array indicating whether each row in
                `arr` is a duplicate.
            int: Number of duplicate rows in `arr`.
    """
    row_sums = np.sum(arr, axis=1)
    if len(np.unique(row_sums)) >= len(arr) * lenient_percent:
        return np.array([]), 0

    if strict:
        duplicated_inds = pd.DataFrame(arr).duplicated()
    else:
        duplicated_inds = pd.Series(row_sums).duplicated()

    return duplicated_inds, np.sum(duplicated_inds)

def round_to_1(x: int | float):# -> float:
    """Round a number to the nearest power of 10.

    Args:
        x (int | float): The number to round.

    Returns:
        float: The rounded number.
    """
    return round(x, -int(np.floor(np.log10(abs(x)))))

def seq_array(size: int, max_val: int = 2000):# -> np.ndarray:
    """Creates an evenly-spaced integer numpy array.

    Args:
        size (int): Length of desired output.
        max_val (int, optional): Desired frequency of output.
            Defaults to 2000.

    Returns:
        numpy.ndarray: An evenly-spaced integer numpy array.
    """
    freq = int(size / max_val)
    time_seq = int(np.ceil(size / freq))
    time_seq = np.repeat([i + 1 for i in range(time_seq)], freq)
    return time_seq[:size]

def order_value_bins(
    vlist: list | np.ndarray | pd.Series,
    segments: list | np.ndarray | pd.Series,
    percentile_diff: float = 0.9
):# -> np.ndarray:
    """Order segments by their geometric means and quantiles.

    Args:
        vlist (list | numpy.ndarray | pandas.Series): 2D array to be sorted.
        segments (list | numpy.ndarray | pandas.Series): Segment label of values.
        percentile_diff (float, optional): Percentile to be evaluated while
            sorting. For example, if set to 0.9, we evaluate the 10th and 90th
            percentiles. Defaults to 0.9.

    Returns:
        numpy.ndarray: list of sorted indices.
    """
    d = pd.DataFrame()
    value_len = 1
    if isinstance(vlist, np.ndarray):
        if len(vlist.shape) == 1:
            vlist = pd.Series(vlist)
        elif len(vlist.shape) == 2:
            vlist = pd.DataFrame(vlist)
        else:
            raise ValueError('Invalid input. Please provide a 1 or 2d array.')
    if isinstance(vlist, list):
        value_len = len(vlist)
        for vi in range(len(vlist)):
            d['v{}'.format(vi)] = vlist[vi]
    elif isinstance(vlist, pd.DataFrame):
        value_len = len(vlist.columns)
        for vi in range(len(vlist.columns)):
            d['v{}'.format(vi)] = vlist[vlist.columns[vi]]
    elif isinstance(vlist, pd.Series):
        d['v0'] = vlist
    else:
        raise ValueError('Invalid input. Please provide a 1 or 2d array.')
    d.index = range(len(d)) # pandas version reconciliation
    if len(segments) != len(d):
        raise ValueError('Segment and value length differ.')

    d['segments'] = segments
    if len(np.unique(segments)) == 1:
        return np.array(range(len(segments)))

    # order segments by their mean
    dg = d.groupby('segments')
    seg_means_all = np.zeros((value_len, len(dg)))
    for vi in range(value_len):
        xi = 0
        for x in dg.groups:
            dgv = dg.get_group(x)['v{}'.format(vi)]
            neg_diff = min(np.min(dgv), 0)
            gm = gmean(dgv - neg_diff) + neg_diff
            q1 = np.quantile(dgv, q=1-percentile_diff) /2
            q2 = np.quantile(dgv, q=percentile_diff) /2
            seg_means_all[vi, xi] = gm + q1 + q2
            xi += 1
    seg_means = np.sum(seg_means_all, axis=0)
    seg_inds = np.array(list(dg.groups.keys()))[np.argsort(seg_means)]

    # rearrange segment indices
    dg_index = np.array([], dtype=int)
    for i in seg_inds:
        dg_index = np.append(dg_index, dg.get_group(i).index)

    return dg_index

def get_timestep(meta: dict, timestep_key: str = '$TIMESTEP'):# -> float | None:
    """Extract timestep from flow cyotometry standard file meta data.

    Args:
        meta (dict): Flow cytometr standard file meta data.
        timestep_key (str, optional): Timestep key in meta data.
            Defaults to '$TIMESTEP'.

    Returns:
        float | None: Timestep if it exists, otherwise None.
    """
    if timestep_key in meta.keys():
        return float(meta[timestep_key])
    return None

## FCS preprocessing functions ####
# Assign column names to index (from fcmdata_helpers).
def _clean_spil_index(spil_df):
    key = spil_df.index
    value = spil_df.columns
    dico = dict(zip(key, value))
    spil_df = spil_df.rename(dico, axis='index')
    return spil_df

# Clean spillover names according to some known formats (from fcmdata_helpers).
def _clean_spil_names(spil_names, keep_left=True):
    new_names = {}
    for name in spil_names:
        # Replace eventual flowJo addition
        if isinstance(name, str) and ' :: ' in name:
            if keep_left:
                # Keep only first part
                new_names[name] = name[:name.find(' :: ')]
            else:
                # Keep second part
                new_names[name] = name[name.find(' :: ') + 4:]
    return new_names

# Apply cleaning on spillover matrix (from fcmdata_helpers).
def _clean_spillover(spillover, data_col=None, keep_left=True):
    if spillover.shape[0] != spillover.shape[1]:
        raise Exception('Error spillover matrix: spillover should be square')

    if np.sum(np.diag(spillover)) != spillover.shape[0]:
        raise Exception('Error spillover matrix: diag should be 1')

    clean_columns = _clean_spil_names(spillover.columns, keep_left)

    spillover = spillover.rename(
        columns=clean_columns
    )
    spillover = _clean_spil_index(spillover)

    if data_col is not None:
        # clean spillover according to data
        intersection = [list(spillover.columns).index(val)
                        for val in data_col if val in list(spillover.columns)]
        if len(intersection) == len(spillover.columns):
            # Reorder spillover matrix
            spillover = spillover.iloc[intersection, intersection]
        else:
            raise Exception(
                "Spillover matrix error: columns don't match with data.")

    # check duplicates
    chan_corrected = copy.copy(list(spillover.columns))
    chan_to_keep = copy.copy(list(spillover.columns))
    for ii in range(len(chan_to_keep)):
        for jj in range(ii + 1, len(chan_to_keep)):
            if sum(abs(spillover[chan_to_keep[ii]] -
                       spillover[chan_to_keep[jj]]) < 1e-10) == len(spillover):
                if chan_to_keep[jj] in chan_corrected:
                    chan_corrected.remove(chan_to_keep[jj])

    return spillover.loc[chan_corrected, chan_corrected]

def get_spillover_raw(meta: dict, dat_columns: list): #list(dat.columns)
    """Extract spillover matrix from fcs meta data.

    Args:
        meta (dict): FCS meta data.
        dat_columns (list): Column names of data.

    Returns:
        pandas.DataFrame: Spillover matrix.
    """
    try:
        s = meta[get_close_match('spill', list(meta.keys()))].split(',')
        n = int(s[0])
        sd = align_list(list1=dat_columns, list2=s[1:(n+1)])  # just in case
        sm = np.array(s[(n+1):]).astype('float64').reshape(n, n)
        n = len(sm)

        # correct diagonal of spill matrix
        smid = True
        for smi in range(sm.shape[0]):
            if sm[smi,smi] != 1.0:
                smid = False
                break
        if smid:
            sm += np.eye(sm.shape[0]) - np.diag(np.diag(sm))

        return pd.DataFrame(sm, columns=sd)
    except:
        warnings.warn("No spillover matrix found.")
        return None

def apply_compensation_matrix(
    data: pd.DataFrame,
    spillover: pd.DataFrame | None
):
    """Apply spillover compensation on data (from fcmdata_helpers).

    Spillover columns should match some subset of data columns.

    Args:
        data (pandas.DataFrame): An event x feature matrix.
        spillover (pandas.DataFrame): spillover matrix to compensate data.

    Returns:
        pandas.DataFrame: Compensated data for columns specified in spillover.
    """
    # TODO: do check on spillover to ensure spillover can be properly applied
    # See https://onlinelibrary.wiley.com/doi/full/10.1002/cyto.a.22018
    if spillover is None:
        return data

    if not isinstance(spillover, pd.DataFrame):
        raise ValueError('Spillover matrix should be a pandas dataframe')

    # Copy data to not overwrite existing dataframe
    data_copy = data.copy()
    spillover = _clean_spillover(spillover, data_copy.columns)
    try:
        spillover_inv = np.linalg.inv(spillover)
    except BaseException:
        spillover_inv = np.linalg.pinv(spillover)
    # X = X * S^(-1)
    data_copy.loc[:, spillover.columns] = np.dot(
        data_copy.loc[:, spillover.columns], spillover_inv)

    return data_copy



import copy
import numpy as np
import pandas as pd
from sklearn.ensemble import IsolationForest
from typing import Callable
from abc import ABCMeta, abstractmethod

## no inheritance for less obscurity METACLEANFCS
def drop_outliers(
    data: pd.DataFrame,
    outlier_func: Callable,
    drop_and: bool = True
):# -> np.ndarray:
    """Applies outlier removal function to given data.

    Args:
        data (pd.DataFrame): Data from which outlier are detected.
        outlier_func (Callable): Outlier detection function.
        drop_and (bool, optional): If True, sets an object as not an outlier if
            all (AND) and its feature values were not flagged as outliers.
            Otherwise, it is not an outlier if any (OR) or its feature values
            were not flagged as outliers.

    Returns:
        np.ndarray: A boolean 1D array the same size as the number of rows in
            `data` labelling non-outliers as True.
    """
    out_keep = ok = np.full((len(data)), 0)
    data_ = copy.copy(data)
    data_['bin'] = np.array(range(len(data))) + 1 # TODO: legacy code
    for c in data.columns:
        du = data_[['bin', c]].drop_duplicates(
            ignore_index=True).copy(deep=True)
        # note: unscaled, works fine
        train_data = np.array(du[c]).reshape(-1, 1)
        train_data[np.isinf(train_data)] = 0
        train_data[np.isnan(train_data)] = 0
        out, _ = outlier_func.fit_predict(train_data, score=True)
        out_keep = out_keep + np.abs(np.minimum(out, ok))

    out_keep = out_keep == 0 if drop_and else out_keep < len(data.columns)
    return out_keep

class OutlierDetector(metaclass=ABCMeta):
    """Super class for `IsoForestDetector` class.
    """
    def __init__(self,  model, **kwargs):
        self.model = model(**kwargs)

    @abstractmethod
    def fit(self):
        pass

    @abstractmethod
    def predict(self):
        pass

    @abstractmethod
    def fit_predict(self):
        pass

class IsoForestDetector(OutlierDetector):
    """Wrapper for outlier detection model, `sklearn.ensemble.IsolationForest`.
    """
    def __init__(
        self,
        model = IsolationForest,
        contamination: float = 0.01,
        n_estimators: int = 500,
        random_state: int = 123
    ):# -> None:
        setting = {
            "contamination":contamination,
            "n_estimators": n_estimators,
            "random_state": random_state
        }
        super(IsoForestDetector, self).__init__(model, **setting)
        self.contamination = contamination
        self.n_estimators = n_estimators
        self.random_state = random_state

    def fit(self, data):
        self.model.fit(data)

    def predict(self, data: np.ndarray, score: bool = True):
        pred = self.model.predict(data)
        if score:
            score_ = self.model.decision_function(data)
        return pred, score_

    def fit_predict(self, data: np.ndarray, score: bool = True):
        self.fit(data)
        return self.predict(data, score=score)

